/*      PWM Motor Speed Controller - PWM_SpeedCtrl.ino
 
 This sketch for the Arduino UNO uses a 16x2 LCD/Keypad shield and a small
 custom op-amp shield to control a variable-speed DC treadmill motor using
 its stock PWM power supply.  We emulate the OEM treadmill controller by
 generating a TTL signal at 20 Hz with a duty cycle between about 1% (min)
 and 100% (max) on digital pin 3.  Multiple options are runtime-selectable
 to support external controls connected to the ADC inputs to start/stop and
 set motor speed.  Shaft RPM may be read by a pulse sensor driving digital
 pin 2 for real-time display but an external low-pass filter (on the custom
 shield) will often be required to eliminate electrical noise from the
 switching power supply.  If RPM is available, linear cutting speed is also
 displayed in real time for the selected cutting tool or work diameter.
 
 This sketch is free software; you can redistribute it and/or modify it.
 It is released as an example application in the hope that it will be useful,
 but without any warranty of merchantability or fitness for a particular purpose.
 
 Vers   ---Date----   --Author--   Description-----------------------------------
 1.10   29-Oct-2012    R.Waters    Fixed SELECT key lockout in AUTORUN mode,
                                   changed default config via EEPROM clear,
                                   added SELECT key @ reset to clear EEPROM,
                                   scaled ADC to "Vref-minus-1LSB"
 1.09   15-July-2012   R.Waters    Added tachometer scale function.
        16-Jun-2012    R.Waters    Enabled reverse polarity on pot/sw inputs,
                                   Added variable step size on auto-repeat.
        12-Jun-2012    R.Waters    Moved enum typedefs into SpdCtrlTypes.h
 1.08   31-May-2012    R.Waters    Removed v1.07 mod (not very useful), added
                                   Run/Stop panel switch as "1-key keypad".
 1.07   13-Apr-2012    R.Waters    Added PWM control of tach comparator threshold
 1.06   28-Jan-2012    R.Waters    Added Switch Input, no auto-repeat on SELECT
 1.05   25-Jan-2012    R.Waters    Ready for Smoke Test
 0.00   30-Dec-2011    R.Waters    Development started
 
********************************************************************************/
//
// We use a "#define" compiler directive to control whether to build optional
// debugging code.  This would typically be excluded from the build when
// development is complete to reduce the size of the code (perhaps by 20%).
// You can observe the generated debugging messages on the serial monitor window
// accessible via the IDE Tools menu (or shift-ctrl-M on Windows).
// Control the build by enabling or "commenting out" the following line:
//
// #define  DEBUG_DUMP  // Enable this line to build debug functions
//
//___________________________________________________________________ INCLUDES
//
// The gnu C++ compiler must know all user-defined data types and external
// functions BEFORE they're used, so they're placed in "header" files (*.h)
// and read in first.  Many header files are dependent upon others, and so
// include them as well.  Due to these interdependencies, sequencing can
// sometimes be tricky, and each header must contain logic that prevents it
// from being processed more than once.
//
// The one-stop shop for many commonly used symbols and functions is the
// header file <Arduino.h> which need not be listed explicitly in your code
// because the IDE automatically includes it for you.  This header defines
// some common symbols needed for Arduino I/O and also brings in several of
// the commonly used standard C library header files.  You can simplify
// your code and avoid reinventing the wheel by using what's here.
// #include <Arduino.h> // don't need this line because it's done for us
//
//    S T A N D A R D     C - L A N G U A G E     L I B R A R I E S
//
// One of the standard headers included by <Arduino.h> is <string.h>,
// the C library implementing functions that manipulate "string" variables
// (sequences of printable characters).  In the C/C++ language, each
// character string is terminated by a NUL character (byte value 0).
// We use string functions like strlen(), strcpy(), strcat(), etc.
// #include <string.h> // don't need this line because it's in Arduino.h
//
// Bring in the definitions of min/max numeric values each integer data
// type can hold.  e.g. defines upper limit of signed integer as INT_MAX.
// We must include this header because it's not referenced in Arduino.h
#include <limits.h>
//
//      A R D U I N O - S P E C I F I C    L I B R A R I E S
//
// The LiquidCrystal library works with all LCD displays compatible with the
// Hitachi HD44780 driver, such as the TC1602 LCD (16 char x 2 line).
// <LiquidCrystal.h> defines the programming interface to this library that
// implements support for the 16x2 LCD display on the LCD/Keypad "shield"
// (option card).  The TC1602 display chip is typical of most textual LCDs,
// and compatible with the Hitachi HD44780.  The library supports both 4- and
// 8-bit mode, but our LCD uses only 4 data lines in addition to RS (register
// select) and Enable (write strobe).  Our LCD has the RW (Read/Write*) line
// hard-wired to ground so we're unable to read back the LCD controller status.
// Commands are therefore issued "blind" using estimated delays and coping
// strategies that retry some operations multiple times.
#include <LiquidCrystal.h>
//
// The Atmel ATmega328P microcontroller on the Arduino UNO has 1K bytes of
// EEPROM (electrically erasable programmable read-only memory). This memory is
// slow to write BUT its values persist after power is lost, essentially like a
// tiny solid-state disk drive.  The <EEPROM.h> library provides byte-by-byte
// access to this memory, but must NOT be called from within an interrupt
// service routine.
#include <EEPROM.h>
//
//      P R O J E C T - S P E C I F I C     H E A D E R S
//
// Bring in definitions of the custom data types used in this sketch.  The use
// of "" instead of <> on the included file name directs the compiler to our
// sketch folder instead of the default library folder. These enumerations must
// be in a separate header file to be processed before the compiler encounters
// the (automatically-generated) function prototypes.  We SHOULD place constants
// and structures in a header file too, but for simplicity I have left them in
// the main sketch.
#include "PWM_SpeedCtrlTypes.h" // define our enumerated data types
//_____________________________________________________
//
// Naming conventions for variables invoke passions like politics or religion
// so many folks are sure to express a dislike of my tastes.  (Please be kind!)
// "Hungarian" notation uses a prefix to indicate the general data type.
// Here are some of the conventions I tried to use in this sketch:
//
//    PREFIX    EXAMPLE     DESCRIPTION
//      b       bMetric     boolean (true or false)
//    i or n    nIndex      integer value
//      e       eMode       enumerated value (user-defined integer)
//      us      usTime      microseconds (an integer or long)
//      ms      msTime      milliseconds
//      sz      szMessage   NUL (zero) terminated character string
//      p       pszMsg      Pointer to (address of) a variable
//      m_      m_nValue    global Member of "this" sketch (from C++)
//
// Constants are named using upper case like "MILLIMETERS_PER_INCH".
//_____________________________________________________
//
// Because the compiler processes #define statements by textual replacement,
// enclosing each value in parenthesis can prevent some hard-to-find errors.
//
// ANY change in the NVSettings structure or its members MUST roll
// the version number to ensure that obsolete EEPROM data is not used.
// e.g. Increment VER_MINOR when NVSettings is modified
#define  VER_MAJOR                (1) // Major version of this sketch
#define  VER_MINOR               (10) // Minor version of this sketch
//
// We use a 16-bit integer to hold the firmware "sketch" version number
// such that the major version is in the high byte, and the minor version
// in the low byte.
#define  VERSION    ((VER_MAJOR << 8) | VER_MINOR)
//------------------------------------//-------------------------
// Properties of Arduino UNO CPU (Atmel ATMega328P-PU)
//       FLASH_SIZE           (32768) // # bytes in Program memory
#define  EEPROM_SIZE           (1024) // # bytes in EEPROM
//       SRAM_SIZE             (2048) // # bytes in Static RAM
//------------------------------------//-------------------------
#define  LCDROWS                  (2) // # of rows (lines) in LCD
#define  LCDCOLS                 (16) // # of columns (chars) in LCD
#define  LCD_DFLT_BRT           (200) // default brightness for splash
#define  UPPER_LINE               (0) // Upper line of 2-line LCD
#define  LOWER_LINE               (1) // Lower line of 2-line LCD
//------------------------------------//-------------------------
//  DIGITAL I/O pins                  // pin  0 (in)  USB Rx data
//                                    // pin  1 (out) USB Tx data
#define  TACH_IRQ                 (0) // pin  2 (in)  tach input (trig IRQ 0)
#define  PIN_TACH                 (2) // pin  2 (in)  tach input
#define  PIN_PWM                  (3) // pin  3 (out) PWM motor power supply @ 20 Hz
#define  PIN_DB4                  (4) // pin  4 (out) LCD bit4
#define  PIN_DB5                  (5) // pin  5 (out) LCD bit5
#define  PIN_DB6                  (6) // pin  6 (out) LCD bit6
#define  PIN_DB7                  (7) // pin  7 (out) LCD bit7
#define  PIN_RS                   (8) // pin  8 (out) LCD Reg Select (cmd/data)
#define  PIN_WRT                  (9) // pin  9 (out) LCD Wrt Ena strobe
#define  PIN_LED                 (10) // pin 10 (out) LCD illum (490 Hz PWM)
//       PIN_XXX                 (11) // pin 11 is not used
//       PIN_XXX                 (12) // pin 12 is not used
#define  PIN_MON                 (13) // pin 13 (out) onboard LED (heartbeat)
//------------------------------------//-------------------------
// "VIRTUAL" ANALOG INPUT channels are used where ADC channels 0-5 typically
// are, but invalid channel numbers are used to invoke special handlers.
#define  ADC_SELECT              (-1) // Virtual ADC for switch input from SELECT key
#define  ADC_TIMING              (-1) // Virtual ADC for diagnostic report of CPU loading
#define  ADC_TACHINP             (-2) // Virtual ADC for diagnostic report of tach input
#define  ADC_UNDEFINED           (-9) // Virtual ADC for initialization of diagnostics
#define  ADC_AUTORUN              (0) // Virtual ADC for switch input = always RUN
//
//  ANALOG INPUT pins & parameters are defined here.  
#define  ADC_LCDPAD               (0) // Channel for LCD shield 5-key pad
#define  ADC_MIN                  (0) // Minimum A/D converter channel
#define  ADC_FREE_MIN             (1) // Minimum "free" A/D converter channel
#define  ADC_MAX                  (5) // Maximum A/D converter channel
#define  ADC_KEY_QUIESCENT     (1000) // Threshold of validity for inactive keypad
#define  ADC_FULL_SCALE        (1023) // Full scale reading in raw counts
#define  ADC_FS_MV             (4995) // Full scale reading in milliVolts
//------------------------------------//-------------------------
#define  MTR_CMD_MAX           (1000) // Ea count = 0.1% of Max (e.g. Range 0-1000)
#define  TACH_SCALE_DENOM     (10000) // Tachometer scale factor = nTachScale/10000
#define  RADIX_DEC               (10) // Show most numbers in base-10
#define  RADIX_HEX               (16) // Show memory adrs in hexadecimal
//------------------------------------//-------------------------
#define  INCHES_PER_FOOT         (12) // UOM conversion factor
#define  SIXTEENTHS_PER_INCH     (16) // UOM conversion factor
#define  MILLIMETERS_PER_INCH    (25) // Integer approximation, close enough for us
#define  MILLIMETERS_PER_METER (1000) // UOM conversion factor
#define  MS_PER_SEC            (1000) // Milliseconds per Second
#define  MS_PER_MINUTE        (60000) // Milliseconds per Minute
//
// ############################################################################
//
//    @@@                     @              @
//   @   @                    @              @
//   @      @@@  @@@@   @@@@ @@@ @@@@  @@@@ @@@  @@@@
//   @     @   @ @   @ @      @      @ @   @ @  @
//   @     @   @ @   @  @@@   @   @@@@ @   @ @   @@@
//   @   @ @   @ @   @     @  @  @   @ @   @ @      @
//    @@@   @@@  @   @ @@@@   @@  @@@@ @   @ @@ @@@@
//
// Constant expressions are evaluated by the compiler and NOT by the target
// machine at runtime, thereby improving performance of your sketch.
// Many of the constants define timing constraints associated with our 20-Hz
// software-generated PWM output to the motor power supply.  The period of
// a 20-Hz signal is 50 milliseconds or 50000 microseconds, and the worst case
// timing for us is a 50% duty cycle, or 25-msec pulse width. We budget 90% of
// this available time for the User Interface routines (USEC_UI_MAX = 22.5 msec).
//
const unsigned long USEC_HRTBEAT = 0x40000;// D13 heartbeat mask in uSec
const unsigned long USEC_PERIOD  = 50000;  // Motor PWM freq = 20 Hz, or 50 ms period
const unsigned long USEC_PW_MIN  =    50;  // Minimum NONZERO output = 0.1% duty
const unsigned long USEC_PW_MAX  = USEC_PERIOD;      // Maximum pulse width
const unsigned long USEC_UI_MAX  = USEC_PERIOD*9/20; // Max time for UI execution
const unsigned long MSEC_REPEAT  =   250;  // keypad auto-repeat interval (msec)
const unsigned long PI_NUMER     =   355;  //   numerator of pi as 355/113
const unsigned long PI_DENOM     =   113;  // denominator of pi as 355/113
const          int  REFRESHLOOPS =    10;  // # of loops between display refresh
const          char SPACES16[]   = "                "; // must be width of LCDCOLS
//
// ############################################################################
//
//    @@@@  @        @           @
//   @    @ @        @           @
//   @      @   @@@  @@@@  @@@@  @
//   @  @@@ @  @   @ @   @     @ @
//   @    @ @  @   @ @   @  @@@@ @
//   @    @ @  @   @ @   @ @   @ @
//    @@@@  @   @@@  @@@@   @@@@ @
//
//      @     @                @
//      @@   @@                @
//      @@   @@  @@@   @@@ @@  @@@@   @@@  @@@  @@@@
//      @ @ @ @ @   @  @  @  @ @   @ @   @ @   @
//      @ @ @ @ @@@@@  @  @  @ @   @ @@@@@ @    @@@
//      @  @  @ @      @  @  @ @   @ @     @       @
//      @  @  @  @@@@  @  @  @ @@@@   @@@@ @   @@@@
//______________________________________________________________
//
// Instantiate the LiquidCrystal library & assign the 4-bit interface
// pins used by the LCD/Keypad shield.  The R/W* line is grounded so
// the LCD status cannot be read; all LCD commands are issued blind.
//
LiquidCrystal m_lcd(PIN_RS,  // LCD Register Select (cmd vs data)
                    PIN_WRT, // LCD Write Enable    (data strobe)
                    PIN_DB4, PIN_DB5,
                    PIN_DB6, PIN_DB7); // data bits 4-7
//______________________________________________________________
//
// Define the structure used to hold operator-adjustable settings
// for off-board inputs from a potentiometer and/or deadman switch.
// This struct holds the ADC (Analog-to-Digital Converter) input
// channel and the Run/Stop threshold voltages in ADC counts.
// For a potentiometer input, these thresholds define the range
// for 0% and 100% speed.  For a switch input, nStpCmd and nRunCmd
// hold the ADC threshold beyond which STOP and RUN commands
// are generated.  e.g. When the input falls between these thresholds,
// no switch command is generated.  Normally, nRunCmd is greater
// (more positive voltage) than nStpCmd.  However, reversing these
// thresholds is allowed and will reverse the command voltage range.
//
typedef struct AdcLimits
{
  short nAtoD;   // ADC Channel No (usually 1 to 5)
  int   nStpCmd; // Lower function threshold in ADC counts (0-1023)
  int   nRunCmd; // Upper threshold in ADC counts
};
//_____________________________________________________________________________
//
// The NVSettings structure holds the Non-Volatile Settings comprising all
// user-selectable configuration parameters.  Only one instance of this
// structure exists, but we use a structure to group them and to simplify
// their transfer to and from nonvolatile EEPROM memory.  These settings
// are restored at power-on to preserve the operator's selected preferences.
// NOTE: If any changes are made to this structure, increment VER_MINOR above
// to prevent obsolete (invalid) settings from being used.
//
struct             NVSettings  // Non-Volatile Settings data block
{
  unsigned int     nVersion;   // Firmware version as Major*0x100 + Minor
  int              nCmdLimit;  // Motor Cmd Limit by 0.1% steps (range 0-MTR_CMD_MAX)
  int              nDiameter;  // Tool diameter in mm (SI) or 16ths (SAE)
  int              nAccelRate; // Motor accel rate (0-50) thousandths/loop
  int              nTachCalib; // Tach calib for Xtal freq error (ms/min)
  unsigned int     nTachScale; // Tach scale for mech drive ratio (10K = 1.0000)
  byte             nTachDiv;   // RPM sensor; 0=disabled, else pulses/rev (1-4)
  byte             nIllumLCD;  // LCD illumination (0-255)
  short            nAuxKeyADC; // External keypad ADC unless ADC_LCDPAD (0)
  boolean          bMetric;    // Units of Measure as true=SI or false=SAE
  struct AdcLimits SwConfig;   // Parameters for RUN/STOP command source
  struct AdcLimits PotConfig;  // Parameters for speed control source
}                  m_Settings; // Only one instance of this structure exists
//
// The size of the NVSetting structure defines our block size in the EEPROM.
// Assuming no padding for word alignment between entities, this would be the
// sum of the sizeof() each member of the structure.  In the Arduino Uno,
// the boolean, char, and short are 1 byte each, the int, enum, and word
// are 2 bytes each, and the long, float, and double are 4 bytes each.
//
const int  SETTINGSIZE = sizeof(m_Settings); // this is 29 in ver 1.08
//______________________________________________________________
//
// GLOBAL variables used by various routines are allocated here.
// We use the "m_" prefix (a C++ naming convention for "member" variables)
// so it will be apparent in the code when one of these is referenced.
//
// Parameters used in processing commands and running the user interface
//
UI_MODE        m_eMode;       // Current item on the display
KEY_ID         m_eKey[3];     // GetKeyCmd() recent decoded keystroke FIFO buffer
KEY_ID         m_ePrevSwCmd;  // Most recent BTN_XXXX command from eGetSwitchCmd()
unsigned long  m_msPrevCmd;   // GetKeyCmd() time of last command for auto-repeat
int            m_nDiagADC;    // Which A/D channel to display during diagnostics
int            m_nRPM;        // most recent driven shaft speed in rev/minute
int            m_nSpeed;      // most recent computed cutting speed, SFPM or m/min
int            m_nRefresh;    // counts REFRESHLOOPS to 0 before refresh display
int            m_nRepeats;    // repeated key count for nGetDelta()
short          m_nSlowSkips;  // counts updates to improve low-RPM performance
short          m_nAuxKeyADC;  // buffered copy of m_Settings.nAuxKeyADC
char  m_szBuffer[LCDCOLS+4];  // buffer used to build display strings
//
// Parameters used in running our PWM output command to the motor power supply
//
boolean        m_bRun;        // Current RUN/STOP state from operator input
int            m_nMotorCmd;   // Most recent cmd; 0 to MTR_CMD_MAX (1000)
unsigned long  m_usPWM_hi;    // uSec for soft PWN output to be high (true)
unsigned long  m_usPWM_lo;    // uSec for soft PWN output to be low (false)
unsigned long  m_usPWM_pw;    // period of soft PWM (m_usPWM_hi + m_usPWM_lo)
unsigned long  m_usUITime;    // elapsed time (uSec) taken by UI routines
boolean        m_bDuringHigh; // determines when we take time to run the UI
//
// These variables are used to implement the tachometer by computing the
// speed of the driven shaft (RPM).  Whenever a pulse is received, the
// Interrupt Service Routine increments m_ulPulses and notes the system
// real-time clock in m_ulMicros.  The "volatile" keyword tells the
// compiler that these variables may change at any time (by the ISR),
// and disables the code optimizer for access to these veriables.
//
volatile unsigned long m_ulMicros;  // Updated by ShaftPulseISR only
volatile unsigned long m_ulPulses;
unsigned long          m_ulPrevMicros;  // Updated by ReadRPM
unsigned long          m_ulPrevPulses;
//
// ##########################################################################
//
//    @@@@@
//   @@   @@           @
//   @@               @@
//   @@@       @@@@  @@@@ @@  @@  @@ @@
//    @@@@    @@  @@  @@  @@  @@  @@@ @@
//     @@@@   @@  @@  @@  @@  @@  @@  @@
//       @@@  @@@@@@  @@  @@  @@  @@  @@
//        @@  @@      @@  @@  @@  @@  @@
//   @@   @@  @@  @@  @@  @@ @@@  @@@ @@
//    @@@@@    @@@@    @@  @@ @@  @@ @@
//                                @@
//                                @@
//                                @@
//
// The setup() function is the first of two functions called by the
// Arduino framework.  It performs system coniguration comprised of
// nonrecurring initialization of hardware and/or software items.
//
void setup()
{
  #ifdef DEBUG_DUMP // build DEBUG output code? -----+
  Serial.begin(9600); // initialize serial support   |
  #endif // -----------------------------------------+

  // Initialize the pin used to control the Motor via 20 Hz PWM
  pinMode(     PIN_PWM, OUTPUT);
  digitalWrite(PIN_PWM, LOW   );       // Ensure motor OFF
  pinMode(     PIN_MON, OUTPUT);
  digitalWrite(PIN_MON, LOW   );       // StatusLED OFF
  pinMode(     PIN_LED, OUTPUT);
  analogWrite( PIN_LED, LCD_DFLT_BRT); // Show splash @ 80% brightness

  m_lcd.begin( LCDCOLS, LCDROWS);  // Initialize LCD driver
  ShowFirmwareVersion(VKEY_INIT);  // Show version as splash screen
  
  // It's possible to configure the Aux Keypad to drive the system bonkers.
  // We provide a quick fix for this situation by wiping out all user
  // settings if operator holds down the SELECT key during startup.
  KEY_ID  eKey = Read5KeyPad(ADC_LCDPAD);  // Read LCD keypad only
  if (eKey == KEY_SELECT)   ClearEEPROM(); // Wipe operator settings?

  RestoreSettings(); // Restore operator's settings from EEPROM

  // Hold display of firmware configuration as a "splash screen".  If we're in
  // AUTORUN mode, delay just half a second for the power supply to stabilize
  // and accept our commands.  Otherwise, we will start the motor via manual
  // command, so there is plenty of time.  Leave the LCD on fairly bright.
  delay((m_Settings.SwConfig.nAtoD == ADC_AUTORUN) ? 500 : 2000);

  analogWrite(PIN_LED, m_Settings.nIllumLCD); // Dim to normal user setting

  // Initialize misc variables
  m_eMode      = MODE_RUN_DEFAULT;          // normal user interface mode
  m_nRefresh   = REFRESHLOOPS;              // RPM display counter
  m_ePrevSwCmd = BTN_RUN;                   // stale RUN cmd suppresses new one

  //  Initialize keypad debounce buffer
  int n = sizeof(m_eKey)/sizeof(m_eKey[0]); // number of entries to clear
  for (int i=0; i<n; i++)                   // loop thru keystroke buffer
       m_eKey[i] = KEY_NONE;                // and clear each location

  // Initialize the tachometer imput by connecting an Interrupt Service Routine
  // to the event of a rising edge input signal.  The pulse train into Digital
  // Pin 2 then generates Interrupt Request 0 which invokes ShaftPulseISR().
  // This routine tallies the pulse count and logs pulse timing.
  attachInterrupt(TACH_IRQ, ShaftPulseISR, RISING);

  // Initialize all PWM control varbs.
  // ADC_AUTORUN differs from ADC_SELECT only at initial state.
  m_bRun        = (m_Settings.SwConfig.nAtoD == ADC_AUTORUN); // RUN state=T/F
  m_nMotorCmd   = 0;
  m_usPWM_hi    = (micros()/USEC_PERIOD + 1) * USEC_PERIOD; // init soft PWM loop
  m_usPWM_pw    = 0;   // final motor command; 0-50K uSec
  m_bDuringHigh = false;
}

//__________________________________________________________
//
//   @@
//   @@
//   @@
//   @@       @@@@    @@@@   @@ @@
//   @@      @@  @@  @@  @@  @@@ @@
//   @@      @@  @@  @@  @@  @@  @@
//   @@      @@  @@  @@  @@  @@  @@
//   @@      @@  @@  @@  @@  @@  @@
//   @@      @@  @@  @@  @@  @@@ @@
//   @@@@@@   @@@@    @@@@   @@ @@
//                           @@
//                           @@
//                           @@
//
//        +--------------+                    +--------------+
//        |              |                    |              |
//   -----+              +--------------------+              +-------
//
//        |              |   ^--RunUI         |
//        |<-m_usPWM_hi->|                    |
//        |                                   |
//        |<-------------USEC_PERIOD--------->|
//
// The loop() function is the second of two functions called by the
// Arduino framework.  It's invoked repeatedly by the Arduino framework
// after system startup to perform recurring operations comprising the
// primary system tasks.  For us, this is generating the PWM motor
// control signal.  The frequency of the hardware Arduino PWM (Pulse
// Width Modulated) outputs controlled by analogWrite() is fixed at
// about 490 Hz.  Therefore, we must "software" generate the 20-Hz 
// PWM signal to command the motor power supply so that our command
// is indistuingishable from the OEM treadmill control panel.
// 
// We generate the PWM output by repeatedly polling the system real-time
// clock and then outputting HIGH and LOW states at appropriate times. 
// The longer of these intervals is used to break out and run
// the user interface.  This includes polling the A/D inputs for
// commands, changing modes, and writing data to the LCD display.
//
void loop()
{
  // Compute the scheduled times for the rising and falling edges of
  // the "soft" PWM output.  The Motor PWM output should be LOW here.
  m_usPWM_hi += USEC_PERIOD;
  m_usPWM_lo  = m_usPWM_hi + m_usPWM_pw;

  if (!m_bDuringHigh)     // Run UI during LOW state?
      RunUserInterface(); // This takes about 12 milliSec

  unsigned long time = micros();
  while (time < m_usPWM_hi)
         time = micros(); // wait here for pulse width interval

  if (m_usPWM_pw > USEC_PW_MIN)  // Zero PW = COMPLETELY off
    digitalWrite(PIN_PWM, HIGH);

  if (m_bDuringHigh)
      RunUserInterface();

  while (time < m_usPWM_lo)
         time = micros(); // wait here for pulse OFF time
         
  digitalWrite(PIN_PWM, LOW);

  // Display 2 Hz "heartbeat" via onboard LED to show loop is running
  digitalWrite(PIN_MON, (time & USEC_HRTBEAT) ? HIGH : LOW);
}

//____________________________________________________________________________
//
// The motor command is handled in its early stages as an integer 0-1000, which
// is essentially tenths of one percent.  The command summarizes these inputs:
// (1) The RUN/STOP state from the selected command input source(s)
// (2) A maximum value set by user preference m_Settings.nCmdLimit
// (3) An external speed control pot, if one is present and enabled
// (4) A command slew rate set by user preference m_Settings.nAccelRate
// The final command is converted into a pulse width in microseconds,
// assuming a maximum of USEC_PW_MAX for a full-power command of 1000.
//
// Returns:  The pulse width in microseconds to be output on the PWM
//           (Pulse Width Modulated) output to the motor power supply,
//           in the range of 0 (motor OFF) to USEC_PW_MAX (100% power).
//
unsigned long ulGetPW_uSec()
{
  // Combine the RUN/STOP command with the preset command limit
  int nNewCmd = m_bRun ? m_Settings.nCmdLimit : 0;

  // Scale command by off-board speed control potentiometer if present.
  // The integer value 0 is boolean false; all other values are true.
  if (nNewCmd && (m_Settings.PotConfig.nAtoD > ADC_MIN))
  {
    // User settings in PotConfig determine the useful range of the pot.
    int nSpan = m_Settings.PotConfig.nRunCmd -
                m_Settings.PotConfig.nStpCmd;

    // Calculate the operator's command within this useful range.
    int nCounts = analogRead(m_Settings.PotConfig.nAtoD) -
                             m_Settings.PotConfig.nStpCmd;

    // Clamp the command to within the expected span, and also
    //  handle the case where the user has inverted the normal
    // range (where +5V is Min and GND is Max command).
    nCounts = (nSpan > 0) ?
              constrain( nCounts, 0, nSpan ) : // normal 0-5V
              constrain( nCounts, nSpan, 0 );  // reversed 5-0V

    // Scale the preset command limit (1000 max) by the pot command,
    // using 32-bit math to avoid overflow.
    nNewCmd = (int)((long)nCounts * (long)nNewCmd / (long)nSpan);
  }

  // Control accel/decel by limiting how fast we slew the command.
  // A full scale command = 1000 and we update at 20 Hz (50 msec).
  // Example: Accelerating to full speed in 2 seconds is 50% per sec,
  // which is 500 counts in 20 updates, or 25 counts per iteration.
  // Therefore, the value of nDelta is 2%/second/count.
  int nDelta = nNewCmd - m_nMotorCmd; // remaining change (+/-)
  if      ((nDelta > 0) && (nDelta > m_Settings.nAccelRate))
            nDelta = m_Settings.nAccelRate; // limit acceleration
  else if ((nDelta < 0) && ((-1 * nDelta) > m_Settings.nAccelRate))
            nDelta = -1 * m_Settings.nAccelRate; // limit deceleration
  m_nMotorCmd += nDelta; // new command for this 50 ms loop

  // Convert command (0 - 1000) into pulse width (0 - USEC_PW_MAX).
  // Use 32-bit long integer to prevent overflow here:
  return (unsigned long)m_nMotorCmd * USEC_PW_MAX / 1000;
}

//__________________________________________________________________
//
// This routine is called from the main 20 Hz loop when we have time
// available to poll the keypad, make calculations, and update the
// display.  The Up/Down keys are handled here ONLY when the motor
// is OFF, to select the User Interface mode (UI_MODE).  All other
// commands from keys or switches are passed into the handler for
// the currently active UI_MODE.
//
void RunUserInterface()
{
  // Since this routine runs 20 times/sec, eKey is usually KEY_NONE.
  boolean bRepeat; // set TRUE when eKey is an auto-repeated arrow
  KEY_ID  eKey = GetKeyCmd(&bRepeat);  // scan for KEY_XXXX cmds

  // Routines using LEFT/RIGHT arrow keys to change a setting become
  // tedious for large changes (ADC settings) unless we increase the
  // step size when the key is held down continuously (auto-repeat).
  // Here we increase the step size from 1 to 2, 4, 8, 16, etc.
  int  nStep = 1;  // default LEFT/RIGHT step size is 1 (ADC counts)
  if  (eKey != KEY_NONE) // ignore gaps between repeated keys
  {
    const int  ACCEL_COUNT = 8;  // Accelerate every 8th repeated key
    m_nRepeats = bRepeat ? (m_nRepeats+1) : 0;    // reset count unless repeating
    for (int i=m_nRepeats/ACCEL_COUNT; i>0; i--)  // Increase step size only
         nStep *= 2;                              // when user holds down key
  }

  // If no keypad command, poll RUN/STOP, Deadman switch, or Pot.
  if  (eKey == KEY_NONE)
       eKey =  eGetSwitchCmd(); // scan for BTN_XXXX cmds

  #ifdef DEBUG_DUMP //--------------------------------------//
  // Report commands to Serial Monitor window on host PC    //
  if (eKey != KEY_NONE) // report any commands              //
      Dump_KEY_ID("RunUserInterface received cmd ", eKey);  //
  #endif // DEBUG_DUMP -------------------------------------//

  // UP/DOWN arrow keys can change the UI Mode ONLY while STOPPED!
  boolean bModeChg = (eKey == KEY_UP) || (eKey == KEY_DOWN);
  if     (bModeChg && !m_bRun && (m_nMotorCmd == 0))
  {
    m_eMode = eSelectUIMode(eKey, m_eMode);
    eKey    = VKEY_INIT; // tell new mode handler to initialize
  }

  // Invoke the user interface handler for the current UI_MODE.
  // Pass in any keypad commands (Left, Right, Select) for processing
  // and increase nStep above the default value of 1 during auto-repeat.
  // The VKEY_INIT "virtual" key tells each handler that their mode has
  // just been activated so they should initialize the whole display.
  switch (m_eMode)
  {
  case MODE_RUN_DEFAULT:        ShowRunningValues(    eKey);               break;
  case MODE_SET_TOOL_DIAM:      SetToolDiameter(      eKey);               break;
  case MODE_SET_MAX_POWER:      SetMaxPower(          eKey, nStep);        break;
  case MODE_SET_ACCEL_RATE:     SetAccelRate(         eKey);               break;
  case MODE_SET_TACH_PPR:       SetTachDivisor(       eKey);               break;
  case MODE_SET_TACH_CALIB:     SetTachCalibration(   eKey, nStep);        break;
  case MODE_SET_TACH_SCALE:     SetTachScaleFactor(   eKey, nStep);        break;
  case MODE_SET_UOM:            SetUnitsOfMeasure(    eKey);               break;
  case MODE_SET_SW_INPUT:       SetSwitchInput(       eKey);               break;
  case MODE_SET_SW_THRESHOLD_U: SetSwitchThreshold(   eKey, nStep, true ); break;
  case MODE_SET_SW_THRESHOLD_L: SetSwitchThreshold(   eKey, nStep, false); break;
  case MODE_SET_POT_INPUT:      SetPotentiometerInput(eKey);               break;
  case MODE_SET_POT_LIMIT_U:    SetPotentiometerLimit(eKey, nStep, true ); break;
  case MODE_SET_POT_LIMIT_L:    SetPotentiometerLimit(eKey, nStep, false); break;
  case MODE_SET_AUX_KEYPAD:     SetAuxiliaryKeypad(   eKey);               break;
  case MODE_SET_DEFAULTS:       SetFactoryDefaults(   eKey);               break;
  case MODE_SHOW_DIAG:          ShowDiagnostics(      eKey);               break;
  case MODE_SHOW_VERSION:       ShowFirmwareVersion(  eKey);               break;
  case MODE_SET_BACKLIGHT:      SetBacklight(         eKey);               break;
  default:                      FatalErr("Invalid mode: ", (int)m_eMode);
  }
}

//____________________________________________________________________
//
// Select the next or previous mode in the sequence, based on the
// current mode (m_eMode) and an UP/DOWN direction command in eKey.
// This routine implements special handlers for those cases where
// mode selection is NOT simply sequential.  If you add a new UI_MODE
// handler, you may need to modify this routine to achieve proper
// sequencing through the User Interface modes.
//
// eKey       The mode change key command from the user, as either
//            KEY_UP to decrement the mode, or KEY_DOWN to increment.
//
// ePrevMode  The currently active user interface mode (from m_eMode)
//
// returns:   The new UI_MODE (User Interface Mode) for the system.
//            A typical mode handler edits one nonvolatile parameter.
//            Therefore, parameters are selected for modification by
//            choosing the appropriate UI mode.
//
UI_MODE  eSelectUIMode(KEY_ID eKey, UI_MODE ePrevMode)
{
  // The UP key decrements the UI_MODE and the DOWN key increments it.
  int nDelta = (eKey == KEY_UP) ? -1 : 1; // mode step direction
  int nMode  = (int)ePrevMode;            // cast enum to int for math

  while (nDelta != 0) // repeat test until we find a valid UI mode
  {
    nMode  += nDelta; // advance to next mode in user's sequence
    switch (nMode)
    {
    // If we ran off the top of the list, wrap around to the bottom
    case  (MODE_RUN_DEFAULT - 1):    return MODE_SET_BACKLIGHT;
    
    // If we ran off the bottom of the list, wrap around to the top
    case  (MODE_SET_BACKLIGHT + 1):  return MODE_RUN_DEFAULT;
    
    // Supperss speed & tach settings if the tachometer is disabled
    case  MODE_SET_TOOL_DIAM:
    case  MODE_SET_UOM:
    case  MODE_SET_TACH_CALIB:
    case  MODE_SET_TACH_SCALE:     if (m_Settings.nTachDiv > 0)
                                       nDelta = 0; // enable the mode
                                   break;
  
    // Skip STOP/RUN details for external switch or pot if disabled
    case  MODE_SET_SW_THRESHOLD_U:
    case  MODE_SET_SW_THRESHOLD_L: if (m_Settings.SwConfig.nAtoD > ADC_AUTORUN)
                                       nDelta = 0; // enable the mode
                                   break;
                                     
    // Skip pot calibration if off-board speed control is disabled
    case  MODE_SET_POT_LIMIT_U:
    case  MODE_SET_POT_LIMIT_L:    if (m_Settings.PotConfig.nAtoD > ADC_MIN)
                                       nDelta = 0; // enable the mode
                                   break;

    default:                       nDelta = 0; // all other cases = allow mode
                                   break;
    }
  }
  return (UI_MODE)nMode;
}


/* ###########################################################################
//
//      @     @   @@@@   @@@@   @@@@@
//      @@   @@  @    @  @   @  @
//      @@   @@  @    @  @   @  @
//      @ @ @ @  @    @  @   @  @@@@@
//      @ @ @ @  @    @  @   @  @
//      @  @  @  @    @  @   @  @
//      @  @  @   @@@@   @@@@   @@@@@
//
//   @   @    @    @   @  @@@@   @     @@@@@  @@@@    @@@
//   @   @   @ @   @@  @  @   @  @     @      @   @  @   @
//   @   @   @ @   @@  @  @   @  @     @      @   @  @
//   @@@@@  @   @  @ @ @  @   @  @     @@@@@  @@@@    @@@
//   @   @  @@@@@  @  @@  @   @  @     @      @  @       @
//   @   @  @   @  @  @@  @   @  @     @      @  @   @   @
//   @   @ @     @ @   @  @@@@   @@@@  @@@@@  @   @   @@@
//
// _________________________________________________________ TEMPLATE
//
// This is the basic template used for all the mode handler routines.
// The operator's most recent key command is passed in as "eKey".
// A "switch" statement or a series of "if" statements are
// appropriate to handle the key command as needed.
// The LCD display is updated as appropriate before exit.
//
// eKey       The key command from the user.  See comments below.
//
void SetMyParameter(KEY_ID eKey) // KEY_ID enumeration passed in
{
  // The KEY_NONE case is a very common situation because the mode handler
  // is called 20 times per second, much more often than keys are pressed.
  // Unless you need to do simething like update the display, you may have
  // nothing to do, in which case you can simply return immediately.
  if (eKey == KEY_NONE)   return;
  //
  // The VKEY_INIT is a virtual (imaginary) key that is passed in when
  // our mode is first selected.  In most cases, we will show a prompt
  // indicating our mode on the upper line of the LCD.  Depending upon
  // need, we can also display a message on the lower line as well.
  // The LcdPrintLine() routine needs 16 chars (full display width).
  if (eKey == VKEY_INIT) { LcdPrintLine(UPPER_LINE, "My Mode Name    ");
                          return; }
  //
  // The SELECT key is the "do it!" command for your mode, so its exact
  // function is up to you.  If we're in the default (running) mode, we
  // may use it to toggle the run/stop state of the motor.  If we're adjusting
  // a configuration setting, we instead save our settings in EEPROM like so:
  if (eKey == KEY_SELECT) { SaveSettingsToEEPROM();  return; }
  //
  // The arrow keys are also dependent upon our context (active mode).
  // LEFT and RIGHT are generally used to decrement and increment the
  // current configuration setting, while UP and DOWN are generally not
  // available because they are reserved to select the mode.
  if (eKey == KEY_UP)       return; // most routines never see this key
  if (eKey == KEY_DOWN)     return; // most routines never see this key
  if (eKey == KEY_LEFT)   { ReduceMySetting();   return; }
  if (eKey == KEY_RIGHT)  { IncreaseMySetting(); return; }
  
  // The latter part of the routine usually updates the LCD lower line
  // to show the current setting.  You can output your messages to the
  // LCD in pieces, but we do this in memory (m_szBuffer) because its 
  // faster & more flexible.  It's critical, though, to be sure you do
  // not exceed the length of m_szBuffer as you prepare your string.
  // Here's an example that displays nMyValue with units of measure.
  char* pBuf = pGetStrBuffer(3);  // Prepare m_szBuffer, indent 3 spaces
  itoab(pEndOf(pBuf), nMyValue, 4, RADIX_DEC); // show nMyValue as 4-digits
  strcat(pBuf, " units");           // append units of measure
  LcdPrintBuffer(LOWER_LINE, pBuf); // xfer buffer to display
}
*/

//_________________________________________________________ MODE_RUN_DEFAULT
//
// This routine handles the user interface during "normal" operation, when
// none of the other SetXxxxx modes are active.  It is the ONLY one called
// while the motor is running, at which time the arrow keys command a change
// in either maximum speed (LEFT/RIGHT) or tool diameter (UP/DOWN).
// The LCD display shows motor PWM command in percent, as well as optional
// speeds in angular (RPM) and linear (Feet or Meters/min) units of measure.
//
//		----------------     Display Field
//               OFF    STOPPED      Top line when no command
//               20%    1803 RPM     Top line when running
//              1-3/8"   648 fpm     Bot line for SAE w/ tach
//               34 mm   192 m/m     Bot line for SI  w/ tach
//		----------------
//
// eKey    The operator's command from a keypad switch (KEY_XXX) or
//         aux switch/pot (BTN_XXXX).  Usually eKey == KEY_NONE.
//
void ShowRunningValues(KEY_ID eKey)
{
  unsigned long usTime = micros();   // entry time for performance check
  
  // First, set the overall RUN/STOP state.  The aux switch (or pot)
  // generates BTN_XXXX and the keypads generte KEY_XXXX commands.
  if      (eKey == KEY_SELECT)   m_bRun = !m_bRun; // SELECT key Toggles
  else if (eKey == BTN_STOP)     m_bRun = false;   // Stop
  else if (eKey == BTN_RUN )     m_bRun = true;    // Run

  if (m_bRun) // Arrow keys adjust parameters only while RUNNING
  {
    if      (eKey == KEY_UP   )  AdjustToolDiameter(true);
    else if (eKey == KEY_DOWN )  AdjustToolDiameter(false);
    else if (eKey == KEY_LEFT )  AdjustMaxPower(-1); // drop by 1%
    else if (eKey == KEY_RIGHT)  AdjustMaxPower( 1); // increase by 1%
  }
  // Get motor command pulse width for 20 Hz PWM output signal
  m_usPWM_pw = ulGetPW_uSec(); // 50,000 uSec = 100% duty cycle

  // Determine when we have time to poll the UI.  This will be
  // either the HIGH or LOW output state, whichever is LONGER
  m_bDuringHigh = m_usPWM_pw > (USEC_PERIOD >> 1);

  // We cannot update the RPM display at 20 Hz because this would
  // cause problems at lower speeds and make it hard to read.
  // Updating twice per second (every 10th pass) seems about right.
  if ((--m_nRefresh) == 0) // time yet to update RPM?
  {
    m_nRefresh = REFRESHLOOPS; // restart delay timer
    //
    // At very low shaft RPM, we may not always get a tach pulse between
    // each display update.  To keep the display from flashing "STOPPED"
    // we allow an extra few update intervals (m_nSlowSkips).
    int  nRPM = ReadRPM();  // will be 0 if no pulse since last pass
    if ((nRPM > 0) || (--m_nSlowSkips <= 0)) // Shall we update display?
    {
      m_nSlowSkips = 5;                 // restart interval extender
      m_nRPM       = nRPM;              // will be 0 if nTachDiv n/a
      m_nSpeed     = CalcSpeed(m_nRPM); // will be 0 if nDiameter n/a
    }
  }

  // ------------------- LCD TOP LINE ----------------------
  char* pBuf = pGetStrBuffer(0);
  // TOP LEFT 6-character field displays motor command in %
  if ((m_usPWM_pw == 0) && !m_bRun) // When user commands STOP and
    strcpy(pBuf, " OFF  ");         // slew is done, show "OFF"
  else
  {
    itoab(pBuf, m_usPWM_pw/500, 3, RADIX_DEC); // 3 char width
    strcat(pBuf, "%  ");
  }
  // TOP RIGHT 10-character field shows RPM iff tach input enabled
  if (m_Settings.nTachDiv == 0)  // if RPM input is not enabled
    strcat(pBuf, pSpc(10));      // clear remainder of display line
  else if (m_nRPM == 0)          // RPM input is enabled but 0
      strcat(pBuf, "   STOPPED");  // message for 0 RPM
  else                           // Tach input shows shaft is turning
  {                              // 1 space + 5 char width for 32K RPM
    if (m_nRPM < INT_MAX)  itoab(pEndOf(pBuf), m_nRPM, 6, RADIX_DEC);
    else                   strcat(pBuf, " *****");
    strcat(pBuf, " RPM");        // units fill out 10-character field
  }
  LcdPrintBuffer(UPPER_LINE, pBuf); // xfer buffer to display

  // ------------------- LCD BOTTOM LINE -------------------------
  // BOTTOM line shows diameter and linear speeds when available
  if ((m_Settings.nTachDiv  == 0) || // Zero indicates tach input n/a
  (m_Settings.nDiameter == 0))   // Zero indicates diameter is undefined
    strcpy(pBuf, pSpc(16));     // suppress output in either case
  else // else speed and diameter are both known
  { // BOTTOM LEFT 8-char field shows selected tool (or work) diameter
    FormatDiameter(pBuf, 8);         // Format OD into the first 8 characters
    // BOTTOM RIGHT 8-char field has cutting speed in selected UOM
    itoab(pEndOf(pBuf), m_nSpeed, 4, RADIX_DEC);
    strcat(pBuf, m_Settings.bMetric ? " m/m" : " fpm");
  }
  LcdPrintBuffer(LOWER_LINE, pBuf); // xfer buffer to display
  
  // Since we're using a simple software timing loop to generate the PWM
  // waveform to the motor power supply, it's important that we don't take
  // too much time in other tasks (running the user interface).  The most
  // sensitive time is when we're running a 50% duty cycle, because this
  // gives us slightly less than HALF the pulse width to run the UI.
  // USEC_UI_MAX is therefore set at 90% of this time (half the period).
  // We log the maximum elapsed time running the UI and halt the system
  // if we exceed our budget.  How close we are to the limit may be
  // displayed by MODE_SHOW_DIAG for the parameter ADC_TIMING.
  usTime = micros() - usTime;  // elapsed microseconds
  if (m_usUITime < usTime)     // capture Max value
      m_usUITime = usTime;
  if (usTime > USEC_UI_MAX)    // >22.5 milliseconds?
      FatalErr("SlowUI; uS:", (int)usTime); // performance warning!
}

//_______________________________________________________ MODE_SET_TOOL_DIAM
//
// Set the tool diameter used to calculate cutting speed in surface
// fet per minute (SFPM) or meters/minute.  We adjust the step size 
// to be about 10% of the diameter for user convenience. We also attempt
// to make the step size change at intervals that will result in round
// numbers of diameter.  However, changing the UOM will break this logic.
// m_Settings.nDiameter is the parameter set by this routine.
//
// eKey    The user's keypad command, most often KEY_NONE.
//         VKEY_INIT during the first pass this mode is active,
//         KEY_LEFT or KEY_RIGHT to decrement or increment our setting,
//         KEY_SELECT to save the current settings in EEPROM.
//
void SetToolDiameter(KEY_ID eKey)
{
  if      (eKey == KEY_NONE)   return;  // nothing to do on this loop
  if      (eKey == KEY_SELECT) { SaveSettingsToEEPROM(); return; }
  if      (eKey == KEY_LEFT)   AdjustToolDiameter(false); // decrement dia
  else if (eKey == KEY_RIGHT)  AdjustToolDiameter(true);  // increment dia
  else if (eKey == VKEY_INIT)  LcdPrintLine(UPPER_LINE, "Tool Diameter:  ");

  // Display diameter in current UOM on LCD lower line
  char* pBuf = pGetStrBuffer(3);  // Prepare buffer, indent 3 spaces
  if (m_Settings.nDiameter == 0)
    strcat(pBuf, "Unspecified  "); // disable display of cutting speed?
  else
    FormatDiameter(pEndOf(pBuf), 13); // Show diameter & UOM
  LcdPrintBuffer(LOWER_LINE, pBuf); // xfer buffer to display
}
//_____________________
//
// Common code to step the diameter of the cutting tool up or down.
// bIncrease = True to increase diameter, False to decrease it.
//
void AdjustToolDiameter(boolean bIncrease)
{
  int nStep; // Size of adjustment in mm or 16ths

  // Scale magnitude of nStep proportionately to current tool diameter
  for (nStep=1; m_Settings.nDiameter > (nStep << 4); nStep*=2);

  if (!bIncrease)         // If we are DECREASING the diameter,
      nStep = -1 * nStep; // make adjustment negative

  // Constrain diameter to be between 0.062" (1 mm) and 12" (300 mm)
  int nMax = m_Settings.bMetric ? 300 /*mm*/ : 192 /*inches*/;
  m_Settings.nDiameter = constrain(m_Settings.nDiameter + nStep, 0, nMax);
}
//
//_______________________________________________________ MODE_SET_MAX_POWER
//
// Set the maximum power level that we will command for the motor, in %.
// If present, the potentiometer's upper limit will be this value.
// m_Settings.nCmdLimit is the parameter set by this routine.
//
// eKey    The user's keypad command, most often KEY_NONE.
//         VKEY_INIT during the first pass this mode is active,
//         KEY_LEFT or KEY_RIGHT to decrement or increment our setting,
//         KEY_SELECT to save the current settings in EEPROM.
//
// nStep   The number of % to change for each Left/Right
//         arrow command key.  The caller is responsible for using
//         the auto-repeat state to increase this (default = 1).
//
void SetMaxPower(KEY_ID eKey, int nStep)
{
  if      (eKey == KEY_NONE)   return;  // nothing to do on this loop
  if      (eKey == KEY_SELECT) { SaveSettingsToEEPROM();  return; }
  if      (eKey == KEY_LEFT)   AdjustMaxPower(nStep * -1); // Left key decrements
  else if (eKey == KEY_RIGHT)  AdjustMaxPower(nStep);      // Right key increments
  else if (eKey == VKEY_INIT)  LcdPrintLine(UPPER_LINE, "Max Motor Power:");

  char* pBuf = pGetStrBuffer(4);    // Prepare buffer, indent 4 spaces
  itoab(pEndOf(pBuf), m_Settings.nCmdLimit/10, 3, RADIX_DEC); // append value
  strcat(pBuf, "%");                // append percent sign
  LcdPrintBuffer(LOWER_LINE, pBuf); // xfer buffer to display
}
//_____________________
//
// Common code to step the max motor command (m_Settings.nCmdLimit) up or down.
// nIncrement determines the size and direction of the change, in % of full power.
//            This is normally +1 to increase, or -1 to decrease, but may
//            use a larger value than 1 for faster changes.
//
void AdjustMaxPower(int nIncrement)
{
  int nStep = nIncrement * 10;  // scale 

  // The command limit set via menu determines motor speed when no off-board
  // speed control pot is connected.  When a pot is connected, this determines
  // the maximum command resulting from the pot at max.  This parameter is in
  // tenths of a percent.  e.g. A value of MTR_CMD_MAX is a command of 100.0 %.
  // Constrain the upper command limit between 1% and 100%.
  const int nMin = MTR_CMD_MAX / 100;  // lower limit is 1% of full power
  m_Settings.nCmdLimit = constrain(m_Settings.nCmdLimit + nStep, nMin, MTR_CMD_MAX);
}

//_______________________________________________________ MODE_SET_BACKLIGHT
//
// Adjust the brightness of the LCD backlight (nIllumLCD).
// Since this is not frequency sensitive, we can use the built-in PWM
// mechanism that runs at about 490 Hz to create a simple D/A converter.  
// m_Settings.nIllumLCD is the parameter set by this routine.
//
// eKey    The user's keypad command, most often KEY_NONE.
//         VKEY_INIT during the first pass this mode is active,
//         KEY_LEFT or KEY_RIGHT to decrement or increment our setting,
//         KEY_SELECT to save the current settings in EEPROM.
//
void SetBacklight(KEY_ID  eKey)
{
  int      nSign = 0;
  if      (eKey == KEY_NONE)   return;  // nothing to do on this loop
  if      (eKey == KEY_SELECT) { SaveSettingsToEEPROM();  return; }
  if      (eKey == KEY_LEFT)   nSign = -1;  // reduce brightness
  else if (eKey == KEY_RIGHT)  nSign =  1;  // increase brightness
  else if (eKey == VKEY_INIT) LcdPrintLine(UPPER_LINE, "LCD Backlight:  ");

  // Adjust by ~12% of current value, but at least 1.
  // Cast nIllumLCD to an integer while doing the math to prevent overflow
  m_Settings.nIllumLCD = constrain((int)m_Settings.nIllumLCD + 
                       nSign * max((int)m_Settings.nIllumLCD >> 3, 1), 0, 255);

  analogWrite(PIN_LED, m_Settings.nIllumLCD); // update brightness of LED

  char* pBuf = pGetStrBuffer(2); // Prepare buffer, indent 2 spaces
  if (m_Settings.nIllumLCD == 0)
      strcat(pBuf, "OFF");         // backlight set to 0% PWM duty
  else
  {
    int pct = map(m_Settings.nIllumLCD, 0, 255, 0, 100); // map to percent
    itoab(pEndOf(pBuf), pct, 3, RADIX_DEC);          // append 3-dig value
    strcat(pBuf, "%");
  }
  LcdPrintBuffer(LOWER_LINE, pBuf); // xfer buffer to display
}

//______________________________________________________ MODE_SET_ACCEL_RATE
//
// Assign the rate at which the motor ramps up to speed (nAccelRate).
// This is set in units of "thousandths of full power per 50ms loop",
// which boils down such that each count is 2% of FULL power per second.
// Because we ratio accel/decel rate to 100% power, the rate is
// independent of the "max power" setting above.
// m_Settings.nAccelRate is the parameter set by this routine.
//
// Example: A value of 50 counts will be displayed here as 100% per second,
// which by definition ramps up from 0 to 100% in 1 second.  Since
// our PWM loop is running at 50 msec, 20 loops make a second.  20 Loops
// at 50 counts each = 1000, which is our internal value for full power.
//
// eKey    The user's keypad command, most often KEY_NONE.
//         VKEY_INIT during the first pass this mode is active,
//         KEY_LEFT or KEY_RIGHT to decrement or increment our setting,
//         KEY_SELECT to save the current settings in EEPROM.
//
void SetAccelRate(KEY_ID  eKey)
{
  int nDelta = 0;
  if      (eKey == KEY_NONE)   return;  // nothing to do on this loop
  if      (eKey == KEY_SELECT) { SaveSettingsToEEPROM();  return; }
  if      (eKey == KEY_LEFT)   nDelta = -1;  // Left key decrements rate
  else if (eKey == KEY_RIGHT)  nDelta =  1;  // Right key increments
  else if (eKey == VKEY_INIT)  LcdPrintLine(UPPER_LINE, "Motor Accel Rate");

  // Constrain motor accel between these limits
  // -   Very slow;  10%/sec  10-sec start    5 thousandths per 50 ms loop
  // - Pretty fast; 100%/sec   1-sec start   50 thousandths per 50 ms loop
  m_Settings.nAccelRate = constrain(m_Settings.nAccelRate + nDelta, 5, 50);

  // Convert value 5-50 into 10-100% for display by shifting left 1 bit
  char* pBuf = pGetStrBuffer(2); // Prepare buffer, indent 2 spaces
  itoab(pEndOf(pBuf), m_Settings.nAccelRate << 1, 3, RADIX_DEC);
  strcat(pBuf, "% per sec");
  LcdPrintBuffer(LOWER_LINE, pBuf); // xfer buffer to display
}

//_________________________________________________________ MODE_SET_TACH_PPR
//
// Assign the tachometer input characteristics in pulses per revolution.
// This is normally 1, especially for a magnetic reed switch sensor.
// However, more pulses per revolution are easy to get with an optical
// sensor and would improve low-speed performance.
//
// m_Settings.nTachDiv is the parameter set by this routine.  A value of
//                     0 disables the tach, otherwise it has the number
//                     of pulses/revolution received.
//
// eKey    The user's keypad command, most often KEY_NONE.
//         VKEY_INIT during the first pass this mode is active,
//         KEY_LEFT or KEY_RIGHT to decrement or increment our setting,
//         KEY_SELECT to save the current settings in EEPROM.
//
void SetTachDivisor(KEY_ID  eKey)
{
  int nDelta = 0;
  if      (eKey == KEY_NONE)   return;  // nothing to do on this loop
  if      (eKey == KEY_SELECT) { SaveSettingsToEEPROM();  return; }
  if      (eKey == KEY_LEFT)   nDelta = -1;  // Left key decrements
  else if (eKey == KEY_RIGHT)  nDelta =  1;  // Right key increments
  else if (eKey == VKEY_INIT)  LcdPrintLine(UPPER_LINE, "Tachometer Input");

  // Constrain tachometer divisor within 1-4 pulses/rev,
  // or 0 to disable the tach and cutting speed functions.
  m_Settings.nTachDiv = constrain(m_Settings.nTachDiv + nDelta, 0, 4);

  char* pBuf = pGetStrBuffer(2); // Prepare buffer, indent 2 spaces
  if (m_Settings.nTachDiv == 0)
      strcat(pBuf, "disabled"); // 0 pulses/rev means tach is N/A
  else
  {
    itoab(pEndOf(pBuf), m_Settings.nTachDiv, 1, RADIX_DEC);
    strcat(pBuf, " pulse"); // make plural for good grammar
    if (m_Settings.nTachDiv > 1) strcat(pBuf, "s");
    strcat(pBuf, "/rev");
  }
  LcdPrintBuffer(LOWER_LINE, pBuf); // xfer buffer to display
}

//_________________________________________________________ MODE_SET_TACH_CALIB
//
// Set the tachometer calibration factor m_Settings.nTachCalib.
// This allows us to compensate for inaccuracy of the system clock
// in our tachometer measurement. This value is used to FINE-tune
// the tachometer by modifying a large constant in the equation.
// We tweak the number of milliseconds per minute for this purpose.
// m_Settings.nTachCalib is the parameter set by this routine.
//
// Example: a calibration value of +60 will be added to the nominal
// value of 60,000, and the result 60060 will be used in computing RPM.
// This will give a +0.1% adjustment in the indicated RPM.
// The easiest way to calibrate this is to use the power line
// frequency as a standard (e.g. 60 Hz = 3600 RPM).
//
// eKey    The user's keypad command, most often KEY_NONE.
//         VKEY_INIT during the first pass this mode is active,
//         KEY_LEFT or KEY_RIGHT to decrement or increment our setting,
//         KEY_SELECT to save the current settings in EEPROM.
//
// nStep   The size of the change for each Left/Right arrow command key.
//         This value is nominally 1 but increases when the inc/dec key
//         is held down.
//
void SetTachCalibration(KEY_ID eKey, int nStep)
{
  // Process the operator's key press into varb "nDelta"
  int     nDelta = 0;
  if      (eKey == KEY_NONE)   return;      // nothing to do?
  if      (eKey == KEY_SELECT) { SaveSettingsToEEPROM();  return; }
  if      (eKey == KEY_LEFT)   nDelta = -1 * nStep; // Left key decrements
  else if (eKey == KEY_RIGHT)  nDelta =      nStep; // Right key increments
  else if (eKey == VKEY_INIT)  LcdPrintLine(UPPER_LINE, "Tach Calibration");

  // Limit max adjustment to 1% of range (1% of 60000 = 600 counts)
  const int MAX_CAL =(long)MS_PER_MINUTE / 100L;
  const int MIN_CAL =(long)MAX_CAL       *  -1L;
  m_Settings.nTachCalib = (int) constrain(m_Settings.nTachCalib + nDelta,
                          MIN_CAL, MAX_CAL); // constrain to +/-600
  
  // Display the calibration delta as a percentage of the nominal value.
  // which would be   Cal% = 100 * nTachCalib / MS_PER_MINUTE.
  // Since this is a small number (a multiple of 0.001667) we do the
  // math using integers, scaled up by 10,000.
  int nPct10K = (long)m_Settings.nTachCalib * 100L / 6L;
  
  // Display the value with decimal shifted to divide by 10K
  char* pBuf = pGetStrBuffer(3); // Prepare msg buffer
  strcat(pBuf, (nPct10K < 0) ? "-" : " "); // insert sign if neg
  nPct10K = abs(nPct10K); // display value as positive
  itoab(pEndOf(pBuf), nPct10K/10000,  1, RADIX_DEC); // single digit
  strcat(pBuf, ".");
  itoab(pEndOf(pBuf), nPct10K%10000, -4, RADIX_DEC); // 4 digits past decimal
  strcat(pBuf, " %");
  LcdPrintBuffer(LOWER_LINE, pBuf); // show current selection
}


//_________________________________________________________ MODE_SET_TACH_SCALE
//
// Set the tachometer scale factor m_Settings.nTachScale (unsigned int).
// This compensates for differences between the tachometer source
// and shaft of interest when the mechanical drive ratio is not 1-to-1.
// Mounting a tach sensor on the motor may be easier than on the quill, 
// so we enter the drive belt or gear ratio here.  Since we're using
// integer math, the value nTachScale contains a value 10,000 times the
// final scale factor, in the range of 0.0100 to 6.553.
//
// Example: If the tach sensor is on the motor shaft and the drive belt
// system reduces the quill to 1/3 motor speed, we would want a tach
// scale factor of 0.3333 and would therefore set nTachScale to 3,333.
//
// eKey    The user's keypad command, most often KEY_NONE.
//         VKEY_INIT during the first pass this mode is active,
//         KEY_LEFT or KEY_RIGHT to decrement or increment our setting,
//         KEY_SELECT to save the current settings in EEPROM.
//
// nStep   The size of the change for each Left/Right arrow command key.
//         This value is nominally 1 but increases when the inc/dec key
//         is held down.
//
void SetTachScaleFactor(KEY_ID  eKey, int nStep)
{
  // Process the operator's key press into varb "nDelta"
  int     nDelta = 0;
  if      (eKey == KEY_NONE)   return;      // nothing to do?
  if      (eKey == KEY_SELECT) { SaveSettingsToEEPROM();  return; }
  if      (eKey == KEY_LEFT)   nDelta = -1 * nStep;  // Left key decrements
  else if (eKey == KEY_RIGHT)  nDelta =      nStep;  // Right key increments
  else if (eKey == VKEY_INIT)  LcdPrintLine(UPPER_LINE, "Tach ScaleFactor");

  // Limit adjustment to the range 0.01 (min) to  6.5535 (max)
  const long MAX_CAL = (long)UINT_MAX; // 6.5535 = 65535/10000
  const long MIN_CAL = 100L;           // 0.0100 =   100/10000
  m_Settings.nTachScale = (unsigned int) constrain(
                          (long)m_Settings.nTachScale + nDelta,
                          MIN_CAL, MAX_CAL);
  
  // Display the scale factor as a floating point number,
  // with decimal shifted to divide by 10K
  char* pBuf = pGetStrBuffer(3); // Prepare msg buffer
  itoab(pEndOf(pBuf), m_Settings.nTachScale/TACH_SCALE_DENOM,  1, RADIX_DEC); // single digit
  strcat(pBuf, ".");
  itoab(pEndOf(pBuf), m_Settings.nTachScale%TACH_SCALE_DENOM, -4, RADIX_DEC); // 4 digits past decimal
  LcdPrintBuffer(LOWER_LINE, pBuf); // show current selection
}


//____________________________________________________ MODE_SET_SW_INPUT
//
// Assign the source of optional motor START/STOP commands based on the
// options defined in the (AdcLimits structure) SwConfig.  e.g. 
// m_Settings.SwConfig.nAtoD is the parameter set by this routine,
// and specifies the ADC channel to receive input from one or two
// switches or a foot-operated potentiometer.
// Commands from this source will be BTN_STOP or BTN_RUN.
//
// eKey    The user's keypad command, most often KEY_NONE.
//         VKEY_INIT during the first pass this mode is active,
//         KEY_LEFT or KEY_RIGHT to decrement or increment our setting,
//         KEY_SELECT to save the current settings in EEPROM.
//
void SetSwitchInput(KEY_ID  eKey)
{
  // Process the operator's key press into varb "nDelta"
  int     nDelta = 0;
  if      (eKey == KEY_NONE)   return;      // nothing to do?
  if      (eKey == KEY_SELECT) { SaveSettingsToEEPROM();  return; }
  if      (eKey == KEY_LEFT)   nDelta = -1;  // Left key decrements
  else if (eKey == KEY_RIGHT)  nDelta =  1;  // Right key increments
  else if (eKey == VKEY_INIT)  LcdPrintLine(UPPER_LINE, "Run/Stop Cmd:   ");

  // Update the selected ADC channel within the range -1 to 5
  // We use -1 and 0 as special cases since these ADCs are n/a for this.
  m_Settings.SwConfig.nAtoD = constrain(m_Settings.SwConfig.nAtoD + nDelta,
                                        ADC_SELECT, ADC_MAX);

  // Display the selected switch input source as an arcane message
  char*   pBuf = pGetStrBuffer(0); // Prepare msg buffer
  switch (m_Settings.SwConfig.nAtoD)
  {
  case ADC_SELECT:                    // ADC Ch -1 initializes to OFF
    strcpy(pBuf, "SELECT key toggl"); // SELECT key toggles RUN/STOP
    break;

  case ADC_AUTORUN:                   // ADC Ch 0 initializes to RUN
    strcpy(pBuf, "Power-On AUTORUN"); // SELECT key toggles RUN/STOP
    break;

  default:          
    strcpy(pBuf, "Pot/Sw on ADC");      // ADC Ch switch/pot for RUN/STOP
    itoab(pEndOf(pBuf), m_Settings.SwConfig.nAtoD, 2, RADIX_DEC);
    break;
  }
  LcdPrintBuffer(LOWER_LINE, pBuf); // show current selection
}

//________________________________________________ MODE_SET_SW_THRESHOLD_X
//
// Assign the values for motor RUN & STOP commands in
// m_Settings.SwConfig.nStpCmd and nRunCmd, respectively.
// The default configuration is for the RUN command threshold to be
// near +4V, and the STOP threshold near +1V, although the correct
// settings depend upon the nature of the external circuitry.
// See the function header for eGetSwitchCmd() for comments on this.
// The gap between is a hysteresis band where no command is issued.  
// The polarity can be reversed by transposing the settings,
// thereby making this input "active low". 
// 
// eKey    The user's keypad command, most often KEY_NONE.
//         VKEY_INIT during the first pass this mode is active,
//         KEY_LEFT or KEY_RIGHT to decrement or increment our setting,
//         KEY_SELECT to save the current settings in EEPROM.
//
// nStep   The number of ADC counts to change for each Left/Right
//         arrow command key.  The caller is responsible for using
//         the auto-repeat state to increase this (default = 1).
//
// bRunCmd Selects which Switch threshold is to be set
//         true modifies the Run threshold, false modifies Stop
//
void SetSwitchThreshold(KEY_ID  eKey,
                        int     nStep,   // size for LEFT/RIGHT step
                        boolean bRunCmd) // False = STOP threshold, True = RUN threshold
{
  const int MIN_GAP = 10; // min ADC counts between Run & Stop
  
  int     nDelta = 0;
  if      (eKey == KEY_NONE)   return;  // nothing to do?
  if      (eKey == KEY_SELECT) { SaveSettingsToEEPROM();  return; }
  if      (eKey == KEY_LEFT)   nDelta = nStep * -1;  // decrease threshold
  else if (eKey == KEY_RIGHT)  nDelta = nStep;       // increase threshold
  // No VKEY_INIT handler exists because UPPER_LINE text is always done

  // Using a pointer to the integer we're modifying simplifies our code
  int* pnCmd = bRunCmd ? &m_Settings.SwConfig.nRunCmd :
                         &m_Settings.SwConfig.nStpCmd;
  *pnCmd = constrain(*pnCmd + nDelta, 0, ADC_FULL_SCALE);

  // Calc the hysteresis band (gap) between the Run & Stop thresholds.
  // The sign indicates whether we've transposed their normal positions.
  int nGap = m_Settings.SwConfig.nRunCmd - m_Settings.SwConfig.nStpCmd;
  
  // -------------- Refresh the LCD Upper Line, like "RUN cmd above:  "
  // Normally, it states whether the Run/Stop command is active hi or low.
  // However,  if these thresholds are too close, we display a warning.
  char*  pBuf = pGetStrBuffer(0);    // Prepare msg buffer
  strcat(pBuf, bRunCmd ? "RUN" : "STOP");
  strcat(pBuf, " cmd ");
  const char*  pText = "gap<min"; // Warning: run/stop gap below minimum
  if (abs(nGap) > MIN_GAP) // Apologies for this arcane logic
         pText = (bRunCmd ^ (nGap<0)) ? "above:" : "below:";
  strcat(pBuf, pText);
  LcdPrintBuffer(UPPER_LINE, pBuf); // xfer buffer to upper line

  // -------------- Refresh the LCD Lower Line
  // Display ADC counts and equiv voltage like "1012  (4.936 V) "
  pBuf = pGetStrBuffer(0);           // Clean old msg buffer
  itoab(pBuf, *pnCmd, 4, RADIX_DEC); // show raw ADC counts
  strcat(pBuf, pSpc(3));
  DisplayAdcVolts(pBuf, *pnCmd);     // Show threshold in VDC
}

//_______________________________________________________ MODE_SET_POT_INPUT
//
// Assign the source of analog motor speed control commands
// in m_Settings.PotConfig.nAtoD.  This mmeber holds the ADC
// channel number (range 1 to 5) when active, or the channel 0
// when no external ADC is in use for speed control.  We use
// ch.0 to mean "none" here as it's hard-wired to the LCD keypad.
//
// eKey    The user's keypad command, most often KEY_NONE.
//         VKEY_INIT during the first pass this mode is active,
//         KEY_LEFT or KEY_RIGHT to decrement or increment our setting,
//         KEY_SELECT to save the current settings in EEPROM.
//
void SetPotentiometerInput(KEY_ID  eKey)
{
  // Process the operator's key press into varb "nDelta"
  int     nDelta = 0;
  if      (eKey == KEY_NONE)   return;      // nothing to do?
  if      (eKey == KEY_SELECT) { SaveSettingsToEEPROM();  return; }
  if      (eKey == KEY_LEFT)   nDelta = -1;  // Left key decrements
  else if (eKey == KEY_RIGHT)  nDelta =  1;  // Right key increments
  else if (eKey == VKEY_INIT)  LcdPrintLine(UPPER_LINE, "Speed Ctrl Pot: ");

  // Move to previous/next selected setting in sequence
  m_Settings.PotConfig.nAtoD = constrain(
  m_Settings.PotConfig.nAtoD+nDelta, ADC_MIN, ADC_MAX);

  char*  pBuf = pGetStrBuffer(0);  // Prepare msg buffer
  if (m_Settings.PotConfig.nAtoD == ADC_MIN)// Ch.0 means NO ext pot present
  {
    strcpy(pBuf, " NONE");
  }
  else // We use ADC channel 1 thru 5 to read ext switches
  {
    strcpy(pBuf, "ADC Channel");
    itoab(pEndOf(pBuf), m_Settings.PotConfig.nAtoD, 2, RADIX_DEC);
  }
  LcdPrintBuffer(LOWER_LINE, pBuf); // xfer buffer to upper line
}

//______________________________________________________ MODE_SET_POT_LIMIT_X
//
// Assign the min/max values for analog motor speed control
// commands in m_Settings.PotConfig.nStpCmd and nRunCmd.
// Pot action may be reversed by transposing these settings.
//
// eKey    The user's keypad command, most often KEY_NONE.
//         VKEY_INIT during the first pass this mode is active,
//         KEY_LEFT or KEY_RIGHT to decrement or increment our setting,
//         KEY_SELECT to save the current settings in EEPROM.
//
// nStep   The number of ADC counts to change for each Left/Right
//         arrow command key.  The caller is responsible for using
//         the auto-repeat state to increase this (default = 1).
//
// bRunCmd Selects which Switch threshold is to be set
//         true modifies the Run threshold, false modifies Stop
//
void SetPotentiometerLimit(KEY_ID  eKey,
                           int     nStep,   // size for LEFT/RIGHT step
                           boolean bRunCmd) // F=Lower Lim, T=Upper
{
  const int MIN_GAP = 100; // min ADC counts between Run & Stop

  int     nDelta = 0;
  if      (eKey == KEY_NONE)   return;  // nothing to do?
  if      (eKey == KEY_SELECT) { SaveSettingsToEEPROM();  return; }
  if      (eKey == KEY_LEFT)   nDelta = nStep * -1;  // decrease threshold
  else if (eKey == KEY_RIGHT)  nDelta = nStep;       // increase threshold
  // No VKEY_INIT handler exists because UPPER_LINE text is always refreshed
  
  // Using a pointer to the integer we're modifying simplifies our code
  int* pnCmd = bRunCmd ? &m_Settings.PotConfig.nRunCmd :
                         &m_Settings.PotConfig.nStpCmd;
  *pnCmd = constrain(*pnCmd + nDelta, 0, ADC_FULL_SCALE);

  // Calc the hysteresis band (gap) between the Run & Stop thresholds.
  // The sign indicates whether we've transposed their normal positions.
  int nGap = m_Settings.PotConfig.nRunCmd - m_Settings.PotConfig.nStpCmd;
  
  // -------------- Refresh the LCD Upper Line, like "MAX spd above:  "
  // Normally, it states whether the Run/Stop command is active hi or low.
  // If these thresholds are too close, display warning "span err".
  char*  pBuf = pGetStrBuffer(0);    // Prepare msg buffer
  strcat(pBuf, bRunCmd ? "MAX Sp" : "MIN Sp");
  const char*  pText = "d span err"; // Warning: min/max span below minimum
  if (abs(nGap) > MIN_GAP) // Apologies for this arcane logic
         pText = (bRunCmd ^ (nGap<0)) ? "eed above:" : "eed below:";
  strcat(pBuf, pText);
  LcdPrintBuffer(UPPER_LINE, pBuf);  // xfer buffer to upper line

  // -------------- Refresh the LCD Lower Line
  // Display ADC counts and equiv voltage like "1012  (4.936 V) "
  pBuf = pGetStrBuffer(0);           // Prepare msg buffer
  itoab(pBuf, *pnCmd, 4, RADIX_DEC); // show raw ADC counts
  strcat(pBuf, pSpc(3));
  DisplayAdcVolts(pBuf, *pnCmd);     // Show threshold in VDC
}

//_____________________________________________________ MODE_SET_AUX_KEYPAD
//
// Configure a second set of pushbuttons on ADCx that implement another source
// of input in addition to the LCD/keypad shield on ADC0.  Note that the
// value used comes from the cached copy of m_Settings.nAuxKeyADC that is
// saved in m_nAuxKeyADC.  We do NOT activate the selection (by updating
// m_nAuxKeyADC) until the settings are saved, since an incorrect input
// would disrupt function of the primary keypad.  This troublesome condition
// can still be fixed by holding LCD keypad SELECT at reset.
//
// The 5-key pad option is assumed to be electrically identical to the
// LCD/keypad shield.  The simplified 1-key option (Start/Stop) must be
// a contact closure to ground.  Both present +5V to the ADC when inactive. 
//
// The integer value at m_Settings.nAuxKeyAD determines our configuration:
//    VALUE    DESCRIPTION
//      0      No auxiliary keypad is connected.
//  +1 to +5   A 5-key pad electrically identical to the one on the
//             LCD/keypad shield is connected to ADC 1 to 5 respectively.
//             See comments for Read5KeyPad() below.
//  -1 to -5   A single momentary switch that presents +5V when inactive
//             and GND when activated is connected to ADC 1 to 5.  This
//             is typically a RUN/STOP button, and generates KEY_SELECT.
//             See comments for () below. 
//
// eKey    The user's keypad command, most often KEY_NONE.
//         VKEY_INIT during the first pass this mode is active,
//         KEY_LEFT or KEY_RIGHT to decrement or increment our setting,
//         KEY_SELECT to save the current settings in EEPROM.
//
void SetAuxiliaryKeypad(KEY_ID  eKey)
{
  if      (eKey == KEY_NONE)     return; // nothing to do?
  if      (eKey == KEY_SELECT) { SaveSettingsToEEPROM();  return; }

  // Sequence nAuxKeyADC like:  0, -1, +1, -2, +2, -3, +3, -4, +4, -5, +5
  boolean   bNeg = m_Settings.nAuxKeyADC < ADC_LCDPAD;  // neg = 1-key, pos = 5-key
  if       (eKey == VKEY_INIT)
            LcdPrintLine(UPPER_LINE, "Aux Keypad/Sw:  ");
  else if ((eKey == KEY_LEFT) && (m_Settings.nAuxKeyADC != ADC_LCDPAD)) // lower limit = ADC0
            m_Settings.nAuxKeyADC = (-1 * m_Settings.nAuxKeyADC) - ( bNeg ? 1 : 0);
  else if  (eKey == KEY_RIGHT)
            m_Settings.nAuxKeyADC = (-1 * m_Settings.nAuxKeyADC) - (!bNeg ? 1 : 0);

  // Make sure we save a valid setting, no matter what happens above
  m_Settings.nAuxKeyADC = constrain(m_Settings.nAuxKeyADC, -1 * ADC_MAX, ADC_MAX);

  char* pBuf = pGetStrBuffer(0); // Prepare scratch buffer for message
  if (m_Settings.nAuxKeyADC < ADC_LCDPAD) // Neg ADC indicates SELECT switch
  {
    strcpy(pBuf, "SLCT btn on ADC");  // Display active ADC channel 1-5
    itoab(pEndOf(pBuf), abs(m_Settings.nAuxKeyADC), 1, RADIX_DEC);
  }
  else if (m_Settings.nAuxKeyADC > ADC_LCDPAD) // Pos ADC indicates 5-key pad
  {
    strcpy(pBuf, "Keypad on ADC");  // Display active ADC channel 1-5
    itoab(pEndOf(pBuf), m_Settings.nAuxKeyADC, 1, RADIX_DEC);
  }
  else // Channel 0 reserved for LCD keys, we use this for "NONE"
    strcpy(pBuf, " NONE");
  LcdPrintBuffer(LOWER_LINE, pBuf); // xfer buffer to upper line
}

//_________________________________________________________ MODE_SET_DEFAULTS
//
// Restore the "factory" default settings into EEPROM on command.
// This will disable all optional input sources.
//
// eKey    The user's keypad command, most often KEY_NONE.
//         VKEY_INIT during the first pass this mode is active,
//         KEY_SELECT to restore default settings in the EEPROM.
//
void SetFactoryDefaults(KEY_ID  eKey)
{
  if (eKey == VKEY_INIT) // Prompt operator action
  {
    LcdPrintLine(UPPER_LINE, "Press SELECT to ");
    LcdPrintLine(LOWER_LINE, "restore defaults");
  }
  else if (eKey == KEY_SELECT) // Wipe out user settings
  {
    SetDefaults();
    ClearEEPROM(); // formerly SaveSettingsToEEPROM();
    m_eMode = MODE_RUN_DEFAULT; // restore default mode
  }
}

//____________________________________________________________ MODE_SHOW_DIAG
//
// Diagnostic utility used to report the state of various parameters.
// The user selects the desired function with the LEF & RIGHT keys,
// which sets the ADC converter channel to be displayed (m_nDiagADC).
// Negative channel numbers are used to select non-ADC functions:
//   -2    Tachometer input state and pulse count on Digital Input pin 2
//   -1    Reports Peak CPU loading as a percentage of budgeted time
// 0 to 5  Show Analog input voltage (0-5 VDC) in real time
//
// The SELECT key is used only for ADC voltage functions to
// latch the current channel for a few seconds.  This allows
// display of the actual voltages seen on the keypad ADC ch-0.
//
// eKey    The user's keypad command, most often KEY_NONE.
//         VKEY_INIT during the first pass this mode is active,
//         KEY_LEFT or KEY_RIGHT to decrement or increment our setting,
//         KEY_SELECT to save the current settings in EEPROM.
//
void ShowDiagnostics(KEY_ID  eKey)
{
  if (eKey == VKEY_INIT)    // initialization?
  {
    LcdPrintLine(UPPER_LINE, "Diagnostic Tests");
    LcdPrintLine(LOWER_LINE, "Select  < or >  ");
    m_nDiagADC = ADC_UNDEFINED; // indicate prompt is displayed

    #ifdef DEBUG_DUMP // ------------------------------+
    DumpSettings(); // show settings on debug window   |
    #endif // -----------------------------------------+
    // We could just return here, but falling thru doesn't hurt
  }

  // Process key strokes into ADC channel # selection
  if      (eKey == KEY_LEFT)   m_nDiagADC--;  // Left key decrements Ch #
  else if (eKey == KEY_RIGHT)  m_nDiagADC++;  // Right key increments Ch #

  if (m_nDiagADC == ADC_UNDEFINED)
      return;  // display verbose user prompt until a key is pressed

  // Once the user has begun function selection, limit it to valid choices.
  // This makes the default selection ADC_TACHINP (closest to ADC_UNDEFINED)
  m_nDiagADC = constrain(m_nDiagADC, ADC_TACHINP, ADC_MAX); // -2 to +5

  // Process the user's requested function
  if      (m_nDiagADC == ADC_TACHINP)     // Virtual ADC -2 shows tachometer input
      ShowISRCount();
  else if (m_nDiagADC == ADC_TIMING)      // Virtual ADC -1 shows CPU Loading
      ShowPeakCPULoad();
  else                                    // Real ADC 0-5, dwelling on request
      ShowADCVoltage(KEY_SELECT == eKey); // ADC # is passed in m_nDiagADC
}
//______________________________
//
// Show the total number of ISR pulses received on the tach pin D2
//
void ShowISRCount()
{
  LcdPrintLine(UPPER_LINE, "D2 Tach pulses: "); // title line
  char* pBuf = pGetStrBuffer(0);
  // Concatenate current pin state and pulse count total
  strcpy(pBuf, HIGH==digitalRead(PIN_TACH) ? "HI  " : "LO  ");
  ultoa(m_ulPulses, pEndOf(pBuf), RADIX_DEC);
  LcdPrintBuffer(LOWER_LINE, pBuf); // xfer buffer to LCD lower line
}
//______________________________
//
// Diagnostic module to display the Peak CPU Loading as a percentage
// of the available time.  At present, there is a 10% pad in this
// threshold USEC_UI_MAX.
//
void ShowPeakCPULoad()
{
  LcdPrintLine(UPPER_LINE, "Peak CPU load:  "); // upper line
  char* pBuf     = pGetStrBuffer(0);
  int   nPctLoad = (int)(((m_usUITime * 100L) + USEC_UI_MAX) / USEC_UI_MAX);
  itoab(pEndOf(pBuf), nPctLoad,  3, RADIX_DEC); // UI time in % of limit
  strcat(pBuf, " %");
  LcdPrintBuffer(LOWER_LINE, pBuf); // xfer buffer to LCD lower line
}
//______________________________
//
// If the SELECT key is pressed, we dwell on the currently selected ADC
// for 10 seconds.  Otherwise we show it once and exit immediately.
//
void ShowADCVoltage(boolean bDwell)
{
  // Use "unsigned long" type to fit delay() function
  const unsigned long DWELL_PERIOD  = 10000; // 10 second dwell
  const unsigned long LOOP_INTERVAL = 250; // msec per loop in dwell mode
  const unsigned int  DWELL_LOOPS   = DWELL_PERIOD / LOOP_INTERVAL;
  
  unsigned int nLoops = bDwell ? DWELL_LOOPS : 1; // SELECT latches for 10 sec
  for (int i=nLoops; i>0; i--) // run 40 loops @ .25 sec each
  {
    // Display lower line as "ADC[x] = X.xxx V"
    char*  pBuf = pGetStrBuffer(0);  // Prepare buffer
    strcpy(pBuf, "ADC[");     // Format prefix
    itoab(pEndOf(pBuf), m_nDiagADC, 1, RADIX_DEC); // ADC #
    strcat(pBuf, "] =");      // prefix is now "ADC[x] ="
    DisplayAdcVolts(pBuf, analogRead(m_nDiagADC)); // append VDC & display

    if (!bDwell) // We're selecting the ADC Channel to display
      LcdPrintLine(UPPER_LINE, "Analog Input:   "); // upper line
    else // We're displaying the selected ADC for 10 seconds
    {
      pBuf = pGetStrBuffer(0); // Prepare buffer
      strcpy(pBuf, "Rem: ");          // Format time remaining
      itoab(pEndOf(pBuf), i >> 2, 2, RADIX_DEC); // in seconds
      strcat(pBuf, " sec");
      LcdPrintBuffer(UPPER_LINE, pBuf); // xfer buffer to upper line
      delay(LOOP_INTERVAL); // delay only during dwell monitor loop
    }
  }
}

#ifdef DEBUG_DUMP // compile for debug build only -------|
//                                                       |
// This routine is called on entry to ShowDiagnostics()  |
// to display the soft configuration data on the serial  |
// debug output window.  We output the number of bytes   |
// in each parameter followed by its current value.      |
//                                                       |
void DumpSettings() //                                   |
{ //                                                     |
  int nSize, nStrucSize=0; //                            |
  Serial.println("Contents of m_Settings:"); // Title    |
  //                                                     |
  // Report firmware version as Major*0x100 + Minor      |
  nSize = sizeof(m_Settings.nVersion);  //               |
  nStrucSize += nSize; //                                |
  Serial.print(nSize); //                                |
  Serial.print("          nVersion 0x"); //              |
  Serial.println(m_Settings.nVersion, HEX); // uint      |
  //                                                     |
  // Motor Cmd Limit by 0.1% steps (0 to MTR_CMD_MAX)    |
  nSize = sizeof(m_Settings.nCmdLimit);  //              |
  nStrucSize += nSize; //                                |
  Serial.print(nSize); //                                |
  Serial.print("         nCmdLimit ");  //               |
  Serial.println(m_Settings.nCmdLimit); // int           |
  //                                                     |
  // Tool diameter in mm (bMetric=True) or 16ths (False) |
  nSize = sizeof(m_Settings.nDiameter);  //              |
  nStrucSize += nSize; //                                |
  Serial.print(nSize); //                                |
  Serial.print("         nDiameter ");  //               |
  Serial.println(m_Settings.nDiameter); // int           |
  //                                                     |
  // Motor accel rate (range 0-50) thousandths per loop  |
  nSize = sizeof(m_Settings.nAccelRate);  //             |
  nStrucSize += nSize; //                                |
  Serial.print(nSize); //                                |
  Serial.print("        nAccelRate ");   //              |
  Serial.println(m_Settings.nAccelRate); // int          |
  //                                                     |
  // Tachometer calibration (range +/-600 ms-per-min)    |
  nSize = sizeof(m_Settings.nTachCalib);  //             |
  nStrucSize += nSize; //                                |
  Serial.print(nSize); //                                |
  Serial.print("        nTachCalib ");   //              |
  Serial.println(m_Settings.nTachCalib); // int          |
  //                                                     |
  // Tachometer Scale Factor (10,000 = 1.0000)           |
  nSize = sizeof(m_Settings.nTachScale);  //             |
  nStrucSize += nSize; //                                |
  Serial.print(nSize); //                                |
  Serial.print("        nTachScale ");   //              |
  Serial.println(m_Settings.nTachScale); // int          |
  //                                                     |
  // RPM sensor; 0=disabled, else pulses/rev (range 1-4) |
  nSize = sizeof(m_Settings.nTachDiv);  //               |
  nStrucSize += nSize; //                                |
  Serial.print(nSize); //                                |
  Serial.print("          nTachDiv "); //                |
  Serial.println(m_Settings.nTachDiv); // byte           |
  //                                                     |
  // LCD illumination (range 0-255)                      |
  nSize = sizeof(m_Settings.nIllumLCD);  //              |
  nStrucSize += nSize; //                                |
  Serial.print(nSize); //                                |
  Serial.print("         nIllumLCD ");  //               |
  Serial.println(m_Settings.nIllumLCD); // byte          |
  //                                                     |
  // External keypad ADC unless 0 (ADC_LCDPAD)           |
  nSize = sizeof(m_Settings.nAuxKeyADC);  //             |
  nStrucSize += nSize; //                                |
  Serial.print(nSize); //                                |
  Serial.print("        nAuxKeyADC ");   //              |
  Serial.println(m_Settings.nAuxKeyADC); // short        |
  //                                                     |
  // Units of Measure as true=SI or false=SAE            |
  nSize = sizeof(m_Settings.bMetric);  //                |
  nStrucSize += nSize; //                                |
  Serial.print(nSize); //                                |
  Serial.print("           bMetric "); //                |
  Serial.println(m_Settings.bMetric);  // bool           |
  //                                                     |
  // Foot Switch ADC Channel No (1 to 5) or 0=none       |
  nSize = sizeof(m_Settings.SwConfig.nAtoD);  //         |
  nStrucSize += nSize; //                                |
  Serial.print(nSize); //                                |
  Serial.print("    SwConfig.nAtoD "); //                |
  Serial.println(m_Settings.SwConfig.nAtoD); // short    |
  // Lower function threshold in ADC counts (0-1023)     |
  nSize = sizeof(m_Settings.SwConfig.nStpCmd);  //       |
  nStrucSize += nSize; //                                |
  Serial.print(nSize); //                                |
  Serial.print("   SwConfig.nStpCmd "); //               |
  Serial.println(m_Settings.SwConfig.nStpCmd); // int    |
  // Upper threshold                                     |
  nSize = sizeof(m_Settings.SwConfig.nRunCmd);  //       |
  nStrucSize += nSize; //                                |
  Serial.print(nSize); //                                |
  Serial.print("   SwConfig.nRunCmd "); //               |
  Serial.println(m_Settings.SwConfig.nRunCmd); // int    |
  //                                                     |
  // Speed Control Pot ADC Channel No (1 to 5) or 0=none |
  nSize = sizeof(m_Settings.PotConfig.nAtoD);  //        |
  nStrucSize += nSize; //                                |
  Serial.print(nSize); //                                |
  Serial.print("   PotConfig.nAtoD "); //                |
  Serial.println(m_Settings.PotConfig.nAtoD); // short   |
  // Lower function threshold in ADC counts (0-1023)     |
  nSize = sizeof(m_Settings.PotConfig.nStpCmd);  //      |
  nStrucSize += nSize; //                                |
  Serial.print(nSize); //                                |
  Serial.print("  PotConfig.nStpCmd "); //               |
  Serial.println(m_Settings.PotConfig.nStpCmd); // int   |
  // Upper threshold                                     |
  nSize = sizeof(m_Settings.PotConfig.nRunCmd);  //      |
  nStrucSize += nSize; //                                |
  Serial.print(nSize); //                                |
  Serial.print("  PotConfig.nRunCmd "); //               |
  Serial.println(m_Settings.PotConfig.nRunCmd); // int   |
  // Size of entire structure                            |
  Serial.print(nStrucSize);       //                     |
  Serial.println(" bytes total"); //                     |
  if (nStrucSize != SETTINGSIZE)  //                     |
      Serial.println("** Size summing ERROR **"); //     |
  Serial.println("---------------"); //                  |
} //                                                     |
#endif // end of DEBUG_DUMP conditional block -----------+


//_________________________________________________________ MODE_SHOW_VERSION
//
// This function is called upon startup as a splash screen,
// and thereafter on user command.
//
// eKey    The user's keypad command, most often KEY_NONE.
//         VKEY_INIT during the first pass this mode is active,
//         KEY_SELECT to exit to normal operating mode.
//         All other commands are ignored.
//
void ShowFirmwareVersion(KEY_ID  eKey)
{
  if (eKey == VKEY_INIT)
  {
    const char* pTitle = "PWM Motor Ctrlr ";
    char*       pBuf   = pGetStrBuffer(3); // Prepare buffer
    strcat(pBuf, "Ver ");
    itoab(pEndOf(pBuf), highByte(VERSION), 1, RADIX_DEC); // major
    strcat(pBuf, ".");
    itoab(pEndOf(pBuf), lowByte(VERSION), -2, RADIX_DEC); // minor
    
    #ifdef DEBUG_DUMP // ------------------------------------+
    strcat(pBuf, " dbg");  // Report debug build FYI         |
    Serial.print(pTitle);  // Echo version to debug window   |
    Serial.println(pBuf);  //                                |
    #endif // DEBUG_DUMP ------------------------------------+

    LcdPrintLine(UPPER_LINE, pTitle);
    LcdPrintBuffer(LOWER_LINE, pBuf); // xfer buffer to LCD lower line
  }
  if (eKey == KEY_SELECT)
      m_eMode = MODE_RUN_DEFAULT; // restore default mode
}

//____________________________________________________ MODE_SET_UOM
//
// Toggle the units of measure between SAE and SI (metric).
// The integer value m_Settings.nDiameter is in 16ths of an inch
// when m_Settings.bMetric is true, or millimeters  when bMetric
// is false.  We toggle between these modes and convert the tool
// diameter between the two systems.
//
// eKey    The user's keypad command, most often KEY_NONE.
//         VKEY_INIT during the first pass this mode is active,
//         KEY_LEFT or KEY_RIGHT to decrement or increment our setting,
//         KEY_SELECT to save the current settings in EEPROM.
//
void SetUnitsOfMeasure(KEY_ID  eKey)
{
  if (eKey == KEY_NONE)   return; // nothing to do?
  if (eKey == KEY_SELECT) { SaveSettingsToEEPROM();  return;  }
  if (eKey == VKEY_INIT)  LcdPrintLine(UPPER_LINE, "Units of Measure");

  if ((eKey == KEY_LEFT) || (eKey == KEY_RIGHT)) // either key toggles units
  {
    m_Settings.bMetric = !m_Settings.bMetric;
    // 
    // Preserve the tool diameter as closely as we can.
    // Conversion is done using ineger math, and attempts to round
    // the result to the nearest whole number by adding in half the
    // denominator to the numerator just prior to the final division.
    //
    if (m_Settings.bMetric) // convert SAE (16ths of an inch) to SI (mm)
        m_Settings.nDiameter = (MILLIMETERS_PER_INCH * m_Settings.nDiameter +
                               SIXTEENTHS_PER_INCH  / 2) / SIXTEENTHS_PER_INCH;
    else // Convert SI (mm) to SAE (16ths of an inch)
        m_Settings.nDiameter = (SIXTEENTHS_PER_INCH * m_Settings.nDiameter + 
                               MILLIMETERS_PER_INCH / 2) / MILLIMETERS_PER_INCH;
  }
  LcdPrintLine(LOWER_LINE, m_Settings.bMetric ?
               " SI (mm, m/min) " : " SAE (in, sfpm) ");
}


// ############################################################################
//
//   @@@@@  @@@@@  @@@@   @@@@    @@@@   @     @
//   @      @      @   @  @   @  @    @  @@   @@
//   @      @      @   @  @   @  @    @  @@   @@
//   @@@@@  @@@@@  @@@@   @@@@   @    @  @ @ @ @
//   @      @      @      @  @   @    @  @ @ @ @
//   @      @      @      @  @   @    @  @  @  @
//   @@@@@  @@@@@  @      @   @   @@@@   @  @  @
//
//  This section contains routines used to manage the transfer of runtime
//  parameters in the m_Settings structure to/from the ATmega328's onboard
//  Electrically Erasable Programmable Read-Only Memory (EEPROM).
//
//_______________________________________________________________ WRITE EEPROM
//
// Save the current operator selections in nonvolatile memory to be
// restored at the next power-on cycle by RestoreSettings().  
// This is done by copying an image of the struct m_Settings to EEPROM.
// EEPROM life is finite; Atmel specifies 100,000 write cycles.
// So we don't squander this resource, we only perform a write when
// data changes exist.  Since our data structure is so much smaller
// than the EEPROM, we can distribute data writes throughout the
// entire 512-byte EEPROM to reduce the affect on any one location.
//
void SaveSettingsToEEPROM()
{
  int i, nNew = 0;             // Default EEPROM adrs for new settings
  int    nOld = FindSettingsInEEPROM(); // Locate prev data (-1 if n/a)
  byte*  pRAM = (byte*)&m_Settings;     // access m_Settings struct as byte seq
  digitalWrite(PIN_PWM, LOW);           // Ensure motor is OFF
  m_eMode     = MODE_RUN_DEFAULT;       
  // restore default mode on exit

  // Determine whether we really need to save settings at all
  for (i=0; i<SETTINGSIZE; i++)      // Determine whether changes were made
    if (*(pRAM+i) != EEPROM.read(nOld+i))
      break;
  if (i == SETTINGSIZE)              // No changes?  No need to update EEPROM
    return;                          // Abort now; Nothing to do

  // Reduce wear on the EEPROM by spreading out writes over all available memory
  if (nOld >= 0)                     // if old settings were found OK
    nNew = nOld + SETTINGSIZE;     // locate new settings just after them
  if (nNew > (EEPROM_SIZE-SETTINGSIZE)) // when end of EEPROM is reached,
    nNew = 0;                      // start over at the beginning

  // nNew points to the destination in EEPROM for the "new" m_Settings.
  for (i=0; i<SETTINGSIZE; i++)      // Scan thru structure byte-by-byte
    EEPROM.write(nNew+i, *(pRAM+i)); // Write image of m_Settings to EEPROM
  for (i=0; i<SETTINGSIZE; i++)
    if (*(pRAM+i) != EEPROM.read(nNew+i)) // Verify image copied OK
      FatalErr("EEPROM Vfy @ ", nNew);  // Aargh! Image verify failed

  // nOld points to the old block to be erased (or -1 if none found).
  if (nOld >= 0)
    for (i=0; i<SETTINGSIZE; i++)       // Erase the old m_Settings by
      EEPROM.write(nOld+i, (byte)0xFF); // cleaing ea byte to FF

    // VERY crude & simple test to verify whether all appears well
  nOld = FindSettingsInEEPROM();
  if (nNew != nOld)                     // try to locate new data set
    FatalErr("EEPROM Clr @ ", nNew);    // Aargh! EEPROM erase failed?

  m_nAuxKeyADC = m_Settings.nAuxKeyADC;  // activte Aux keypad setting
  ReportWrite2EEPROM(SETTINGSIZE, nNew, " Settings Saved ");
}

//__________________________________________________________________
//
// Clear all user settings by wiping the EEPROM to all 0xFF.
// This will behave exactly the same as saving the default preferences,
// except that the EEPROM will be wiped totally clean.  This can be
// handy if you want to run another sketch with a clean EEPROM.
//
void ClearEEPROM()
{
  int    nBytes=0;
  int    nLoc=0;

  // Count down so nLoc will end up with block start adrs
  for (int adrs = EEPROM_SIZE-1; adrs >= 0; adrs--)
  {
    if (EEPROM.read(adrs) != 0xFF) // if any data is present
    {
      EEPROM.write(adrs, 0xFF);    // wipe byte to 1's
      nLoc = adrs;                 // track start of data block
      nBytes++;                    // count # of bytes erased
    }
  }
  ReportWrite2EEPROM(nBytes, nLoc, "Settings Cleared");
}

//__________________________________________________________________
//
// Common code shared by routines that write to EEPROM.
// Report the action on LCD and debug output window.
//
// nBytes  The number of bytes actually written
// nLoc    The offset within EEPROM of the start of the block
// pMsg    A 16-character message for the LCD upper line.
//
void ReportWrite2EEPROM(int nBytes, int nLoc, const char* pMsg)
{
  // Report success of EEPROM write to user  
  char* pBuf = pGetStrBuffer(0);  // Prepare empty text buffer
  
  #ifdef DEBUG_DUMP //------ compile for debug build only ------+
  itoab(pBuf, nBytes, 0, RADIX_DEC);        // config data size |
  strcat(pBuf, " bytes @ "); //                                 |
  itoab(pEndOf(pBuf), nLoc, -4, RADIX_HEX); // hex data adrs    |
  strcat(pBuf, "H");     //                                     |
  Serial.print(pMsg);    //                                     |
  Serial.print(": ");    //                                     |
  Serial.println(pBuf);  //                                     |
  #endif //-----------------------------------------------------+
  
  LcdPrintLine(  UPPER_LINE, pMsg); // Report caller's msg on upper line
  LcdPrintBuffer(LOWER_LINE, pBuf); // Report debug data if any
  delay(2000);
}


//_______________________________________________________________ READ EEPROM
//
// Retrieve the operator's desired settings from nonvolatile mamory
// during power-on initialization, placing them in the struct m_Settings.
//
void RestoreSettings()
{
  int nSrc = FindSettingsInEEPROM();
  if (nSrc >= 0) // NVSettings block found?
  {
    // Initialize the m_Settings structure from the saved image in
    // EEPROM by copying it back into m_Settings.  This is done as
    // a byte-by-byte copy of the memory block, similar to memcpy().
    // This is why we use the version number as an identifier, so that
    // we don't use an obsolete structure with incompatible content.
    byte* pDst = (byte*) &m_Settings;
    for (int i=0; i<SETTINGSIZE; i++)
      *(pDst+i) = EEPROM.read(nSrc+i);
      
    m_nAuxKeyADC = m_Settings.nAuxKeyADC;  // activate Aux keypad setting
    return;
  }
  // Set default values if EEPROM image is uninitialized
  // or from a previous version of the firmware.
  SetDefaults();
}

//______________________________________________________________
//
// Initialize the operator's settings to default values.
// These are the values used the first time your Arduino
// runs this sketch, because the EEPROM is empty.  These
// settings are also used whenever the VERSION changes,
// or when the user commands SetFactoryDefaults().
//
void SetDefaults()
{
  m_Settings.nVersion    =          VERSION; // Firmware version = X.xx
  m_Settings.nCmdLimit   =  MTR_CMD_MAX / 2; // Motor Cmd = half-power
  m_Settings.nDiameter   =                8; // Tool diameter = 8/16ths of an inch
  m_Settings.nAccelRate  =           32 / 2; // 32% per sec = 100% cmd in 3 sec
  m_Settings.nIllumLCD   = LCD_DFLT_BRT / 4; // LCD illumination about 20%
  m_Settings.nTachDiv    =                0; // Tachometer input = disabled
  m_Settings.nTachCalib  =                0; // No tachometer tuning for Xtal freq error
  m_Settings.nTachScale  = TACH_SCALE_DENOM; // 10,000 = 1.0000 Scale factor
  m_Settings.bMetric     =            false; // Units of Measure = SAE (inches)
  m_Settings.nAuxKeyADC  =       ADC_LCDPAD; // No secondary keypad/switches
  m_nAuxKeyADC           =       ADC_LCDPAD; // Activate Aux keypad setting

  // ADC outputs 1023 counts in +5V, for 205 counts/volt
  m_Settings.SwConfig.nAtoD    = ADC_SELECT; // NO active switch (use SELECT key)
  m_Settings.SwConfig.nStpCmd  =        200; // <1.0 VDC = STOP cmd
  m_Settings.SwConfig.nRunCmd  =        800; // >4.0 VDC = RUN cmd

  m_Settings.PotConfig.nAtoD   =    ADC_MIN; // NO active speed control Pot
  m_Settings.PotConfig.nStpCmd =         20; // <0.1V commands minimum speed
  m_Settings.PotConfig.nRunCmd =       1020; //  5.0V commands full speed
}

//_________________________________________________________________
//
// The EEPROM in the ATMega328P has a finite life in terms of the number
// of write cycles for each location.  We save our m_Settings in different
// locations each time we write to EEPROM in order to reduce the activity
// on any one location within the EEPROM.  We use the version number
// at the start of the m_Settings structure to ID the block of data.
// This technique reduces the number of saves before reuse of the same
// memory location to EEPROM_SIZE / SETTINGSIZE;
//
// This function returns the 0-based offset within the EEPROM of the
// image of the m_Settings structure.  If the m_Settings were not found,
// this function returns a negative number.
//
int FindSettingsInEEPROM()
{
  for (int adrs=0; adrs < EEPROM_SIZE; adrs += SETTINGSIZE)
  {
    // The ATMega is a "little Endian" machine (like most Intel chips),
    // with the low-order byte stored first in memory.
    if (( lowByte(VERSION) == EEPROM.read(adrs  )) &&
        (highByte(VERSION) == EEPROM.read(adrs+1)) )
           return(adrs);
  }
  #ifdef DEBUG_DUMP // compile for debug build only ------------//
  Serial.println("Nonvolatile settings NOT found in EEPROM");   //
  #endif DEBUG_DUMP // -----------------------------------------//
  return (-1);      // settings not found
}


// ############################################################################
//
//   @   @                             @       @@
//   @  @                              @      @ @
//   @ @    @@@  @   @ @@@@  @@@@   @@@@      @ @
//   @@ @  @   @  @ @  @   @     @ @   @      @@
//   @  @  @@@@@  @ @  @   @  @@@@ @   @     @  @@
//   @   @ @      @ @  @   @ @   @ @   @     @  @
//   @   @  @@@@   @   @@@@   @@@@  @@@@      @@ @
//                 @   @
//               @@    @
//
//         @@@            @  @        @
//        @   @              @        @
//        @     @   @   @ @ @@@  @@@  @@@@   @@@   @@@@
//         @@@   @  @  @  @  @  @   @ @   @ @   @ @
//            @  @ @ @ @  @  @  @     @   @ @@@@@  @@@
//        @   @   @   @   @  @  @   @ @   @ @         @
//         @@@    @   @   @  @@  @@@  @   @  @@@@ @@@@
//___________________________________________________________________
//
// Return the currently active keypad command as a KEY_ID.  Keypads
// are defined as a momentary switch that the user closes "briefly".
// The "deadman" switch that is held closed to maintain motor operation
// is handled elsewhere because the "repeat" logic is quite different.
//
// The LCD keypad shield on ADC0 is ALWAYS scanned.  If enabled in
// the EEPROM configuration settings, other command sources are
// scanned as well.  These can be:
// - A second full keypad that is electrically identical to the one
//   on the LCD keypad, but typically with more robust switches.
// - A single "Select" switch that is active LOW.  This works like
//   the keypad, but has only 1 switch.  Commands RUN/STOP alternately.
//
// pbRepeat contains the address of the  variable to receive a boolean
//          indicating whether or not the returned KEY_ID is an
//          auto-repeated arrow command.  This technique allows a
//          function to return more than one value.
//
KEY_ID GetKeyCmd(boolean* pbRepeat)
{
  // The first half of this routine checks all available sources
  // for a key or switch input command.  These inputs are:
  // (1) The 5-button keypad on the LCD shield (always present)
  // (2) The off-board "aux" keypad if one is present and enabled,
  // (3) The off-board RUN/STOP switch or speed control pot, if enabled.
  KEY_ID  eKey = Read5KeyPad(ADC_LCDPAD); // Read A/D, map to Cmd

  // if no command from LCD 5-key pad, try aux 5-key input if enabled
  if ((eKey == KEY_NONE) && (m_nAuxKeyADC > ADC_LCDPAD)) // ADC > 0
       eKey =  Read5KeyPad(  m_nAuxKeyADC); // Scan Aux 5-key pad

  // if no command from above, try aux single-key pad (start/stop sw)
  if ((eKey == KEY_NONE) && (m_nAuxKeyADC < ADC_LCDPAD)) // ADC < 0
       eKey =  Read1KeyPad(  m_nAuxKeyADC); // Scan Aux Sw
       
  //---------- 2nd half of routine processes key from above ----------
  // Implement contact debounce and auto-repeat for keys held down.
  // The somewhat arcane logic below is broken down into individual
  // statements in an attempt to aid understanding of the logic.
  unsigned long msTime = millis(); // auto-repeat needs current time

  // Maintain FIFO buffer of the last 3 keypad inputs.
  m_eKey[2] = m_eKey[1];  // When new data comes in,
  m_eKey[1] = m_eKey[0];  // ripple the old data thru.
  m_eKey[0] = eKey;

  *pbRepeat = false; // setup aux return parameter for most cases below
 
  // First, check to see that we have 2 identical successive readings
  // in order to suppress transients and switch contact "bounce".
  if (m_eKey[0] != m_eKey[1])   return KEY_NONE;

  // The auto-repeat feature applies when a key switch is held down
  // continuously.  We detect this state as equality in all of m_eKey[].
  if (m_eKey[0] == m_eKey[2])
  {
    // Wait for the repeat interval between issued commands.
    // We do not differentiate between the initial delay at the
    // start of auto-repeat and the delay between subsequent
    // keys, so they are handled the same.
    if (msTime < (m_msPrevCmd + MSEC_REPEAT))  return KEY_NONE;
    
    // Enable the repeat function ONLY for the ARROW keys.
    // The primary reason is to inhibit repeats on the SELECT key
    // because it's used to toggle RUN/STOP mode.  It's also cleaner
    // to inhibit repeats of STOP/RUN from the deadman switch
    // because they are unnecessary.
    if ((m_eKey[0] < KEY_RIGHT) ||
        (m_eKey[0] > KEY_LEFT))      return KEY_NONE;
    
    // If we make it here, we have an auto-repeat ARROW command
    *pbRepeat = true; // Tell caller that cmd is an auto-repeat
  }
  // If we fall thru to here, we have a command to process!
  m_msPrevCmd = msTime; // log current time for next pass thru
  return eKey;
}

//_________________________________________________________________
//
//                                      L C D    5 - K E Y    P A D
//
// The keypad on the LCD shield contains a resistor/switch network
// much like the cruise control on an automobile steering wheel.
// Each switch creates a unique voltage divider ratio, powered by
// the +5V regulator and read by the Channel 0 A/D converter.
// Since the switches on the LCD shield are small and fragile, an
// external keypad using the same circuit may be attached to ADC Ch5.
//
//                 +--------- +5V
//                 |
//               2000 ohms
//                 |           A/D Input
//   +----RIGHT----+---------> Ch.0 default,
//   |             |           Ch.5 optional
//   |            330 ohms
//   |             |
//   +-----UP------+
//   |             |
//   |            620 ohms
//   |             |
//   +----DOWN-----+
//   |             |
//   |           1000 ohms
//   |             |
//   +----LEFT-----+
//   |             |
//   |           3300 ohms
//   |             |
//   +---SELECT----+
//   |
//   +---------------------- GND
//
//  The 10-bit ADC reads from 0V (0 counts) to +5V (1023 counts),
//  depending on which keyswitch is depressed.  "Ideal" design
//  voltages as well as the actual ones measured on my board are:
//
//           ---------------DESIGN------------   ---MEASURED---
//  KEY        Resistor Ratio     VDC   counts   counts   VDC
//  ------   -----------------   -----  ------   ------  -----
//  RIGHT       0/2000 = 0.000   +0.00      0        1   +0.00
//  UP        330/2330 = 0.142   +0.71    145      129   +0.63
//  DOWN      950/2950 = 0.322   +1.61    329      305   +1.49
//  LEFT     1950/3950 = 0.494   +2.47    505      478   +2.34
//  SELECT   5250/7250 = 0.724   +3.62    741      720   +3.52
//  (none)   2000/2000 = 1.000   +5.00   1023     1023   +5.00
//
//  Read the specified A/D converter and map the reading into a KEY_ID
//
KEY_ID  Read5KeyPad(int adcChannel)
{
  // Read the caller's specified A/D converter channel,
  //  (0-5V) as 0-1023 counts (4.883 mV/count)
  unsigned int adcCounts = analogRead(adcChannel);

  // Check the most common condition first for best speed,
  // although performance is not presently a problem.
  // "no key" design value = 1023 counts (anything above 882; > 4.32 V) 
  if (adcCounts > 882)  return KEY_NONE;
  
  // SELECT key design value = 741 counts (range 624-882; 3.05 - 4.31 V) 
  if (adcCounts > 623)  return KEY_SELECT;
  
  // LEFT key design value = 505 counts (range 418-623; 2.04 - 3.04 V) 
  if (adcCounts > 417)  return KEY_LEFT;
  
  // DOWN key design value = 329 counts (range 238-417; 1.16 - 2.04 V) 
  if (adcCounts > 237)  return KEY_DOWN;
  
  // UP key design value = 145 counts (range  73-237; 0.36 - 1.16 V)   
  if (adcCounts >  72)  return KEY_UP;
  
  // RIGHT key design value = 0 counts (anything below 72;  < 0.35 V)
  return KEY_RIGHT;
}

//_________________________________________________________________
//
//                           A U X     S I N G L E - K E Y    P A D
//
// The optional single RUN/STOP switch is an "alternate action"
// command switch that allows a single momentary button to start
// AND stop the motor.  This is connected like the SPST or SPDT
// switches shown below under eGetSwitchCmd() but its momentary
// nature is better suited for a keypad handler like Read5KeyPad().
// We therefore implement a single-key version here.
//
// Read the A/D converter and map the reading into a KEY_ID.
// We accept negative values for the ADC to support preceding logic.
//
KEY_ID  Read1KeyPad(int adcChannel)
{
  // Read the A/D converter (0-5V) as 0-1023 counts (4.883 mV/count)
  unsigned int adcCounts = analogRead(abs(adcChannel));

  // We should see +5V when inactive, or 0V with switch depressed.
  // SPST open has pull-up resistor to Vcc.
  // SPDT "NO" pole connects to Vcc and looks just the same.
  if (adcCounts > 50)  return KEY_NONE;  // anything over 0.24V
  
  return KEY_SELECT;  // < 0.24V = Contact closure to ground = active
}


//_________________________________________________________________
//
// Here are external Switch Schematics compatible with mode SWI_SWITCH
// and handled by eGetSwitchCmd().  In general, a single switch (or pot)
// must be held in the "on" position for the motor to run.  This works
// very well for a foot switch or a foot-operated speed control like a
// sewing machine.  Two switches are required to make a latching system
// where "start" and "stop" are separate commands.
// The polarity (+5V and Gnd) shown below may be reversed by setting 
// m_Settings.SwConfig.nStpCmd more positive than nRunCmd.
// The ADC reads 1023 counts on a +5V input, which is then 205 counts
// per volt, or 4.88 mV per count.
//
// The schematic below depicts a pair of momentary (pushbutton) switches
// connected with resistors and an optional capacitor to generate discrete
// voltages read by the A/D Converter.  The "RUN" switch is normally open,
// and the "STOP" switch is normally closed. The most recent command is 
// latched by software.  The 3 states are:
//  (a) 5.0 VDC when START is pressed and STOP is not,
//  (b) 0.0 VDC when STOP is pressed (START is ignored), and
//  (c) 2.5 VDC when neither switch is pressed.  
// Suggested m_Settings.SwConfig thresholds: Run=4.0V, Stop=1.0V
//
//     +-----------+------> +5V
//     |           |
//  2700_ohms   START_NO
//     |           |
//     +-----------+
//                 |
//               STOP_NC
//                 |
//     +-----------+------> to ADC
//     |           |
//  2700_ohms    0.1_uF
//     |           |
//     +-----------+------> GND
//
//
// This alternate schematic is suitable for a simple SPST switch that
// can either be the conventional latching (ON/OFF) type, or may be
// momentary and held on by the operator as in a "deadman" foot switch.
// Suggested m_Settings.SwConfig thresholds: Run=4.0V, Stop=3.0V
//
//     +------------------> +5V
//     |
//   RUN_NO
//     |
//     +-----------+------> to ADC
//     |           |
//  2700_ohms    0.1_uF (optional)
//     |           |
//     +-----------+------> GND
//
// This alternate schematic eliminates the resistor above and uses
// a SPDT switch instead, which reduces the parts count slightly.
// This is the configuration actually used in my project's hardware.
// Suggested m_Settings.SwConfig thresholds: Run=4.0V, Stop=3.0V
//
//     +------------------> +5V
//     |
//     O
//        _O--------------> to ADC
//     O -
//     |
//     +------------------> GND
//
// Applications like a TIG welder or sewing machine use a foot-operated
// potentiometer to command start/stop as well as provide speed control.
// We can do the same by setting both START and STOP thresholds near the
// lower end of travel.  Suggested m_Settings.SwConfig thresholds in this
// case would be Run=0.2V, Stop=0.1V.
//
//        +---------------> +5V
//        |
//      +-+-+
//      |   |  10K
//      | P |
//      | O |<------------> to ADC
//      | T |
//      |   |
//      +-+-+
//        |
//        +---------------> GND
//
//
// eGetSwitchCmd() returns the key command from an external source.
// This routine supports multiple input configurations from an off-board
// potentiometer or switch(es), such as the ones depicted above.  The input
// voltage is mapped into 3 ranges divided by 2 user-selectable thresholds.
// The middle range always returns command BTN_NONE (no command).
// The outermost two ranges return BTN_RUN and BTN_STOP, and may be
// reversed by setting the thresholds appropriately.
//
// Note: RUN/STOP commands (returned from this routine) are ignored when
//       configuring the system.
//
KEY_ID  eGetSwitchCmd()
{
  KEY_ID  eSwCmd = KEY_NONE; // Init default returned command
  
  // If our ADC channel for switch input is either ADC_SELECT (-1)
  // or ADC_AUTORUN (0) then off-board switches are not used, so
  // we have nothing to do here.
  if (m_Settings.SwConfig.nAtoD > ADC_AUTORUN)
  {
    // Read the voltage from the off-board pot or switch circuit
    int adcCounts = (int)analogRead(m_Settings.SwConfig.nAtoD);

    // Handle the case where input voltage levels are reversed.
    // A "normal" situation is when pot VDC increases directly with command.
    // e.g. A reverse situation is when +5V is min and GND is the max cmd.
    bool bNormal = m_Settings.SwConfig.nRunCmd > m_Settings.SwConfig.nStpCmd;

    // Test the ADC reading against the RUN & STOP command thresholds
    if (bNormal ^ (adcCounts > m_Settings.SwConfig.nStpCmd))  eSwCmd = BTN_STOP;
    if (bNormal ^ (adcCounts < m_Settings.SwConfig.nRunCmd))  eSwCmd = BTN_RUN;

    // Suppress redundant commands from this switch
    if     (m_ePrevSwCmd == eSwCmd)                           eSwCmd = KEY_NONE;
    else    m_ePrevSwCmd  = eSwCmd;
  }
  // If input voltage fails required thresholds, we have no cmd
  return eSwCmd;  // all other cases
}

//______________________________________________________________________________
#ifdef DEBUG_DUMP // enable this function for DEBUG builds only
//
// Diagnostic routine to print a keystroke or button command
//
// pLabel  The label to precede the keystroke in the debug output
// eKey    The enumeratied value of the command to be displayed
//
void Dump_KEY_ID(char* pLabel, KEY_ID eKey)
{
  Serial.print(pLabel);
  Serial.print(" KEY_ID=");
  switch (eKey)
  {
    case  KEY_NONE:   pLabel = "KEY_NONE (0)";   break;
    case  KEY_RIGHT:  pLabel = "KEY_RIGHT (1)";  break;
    case  KEY_UP:     pLabel = "KEY_UP (2)";     break;
    case  KEY_DOWN:   pLabel = "KEY_DOWN (3)";   break;
    case  KEY_LEFT:   pLabel = "KEY_LEFT (4)";   break;
    case  KEY_SELECT: pLabel = "KEY_SELECT (5)"; break;
    case  VKEY_INIT:  pLabel = "KEY_INIT (6)";   break;
    case  BTN_STOP:   pLabel = "BTN_STOP (7)";   break;
    case  BTN_RUN:    pLabel = "BTN_RUN (8)";    break;
    default:          pLabel = "UNDEFINED!";
  }
  Serial.println(pLabel);
}
#endif // DEBUG_DUMP
//________________________________________________________________________________________
//
//   @@@@   @@@@   @     @       @@       @@@   @@@@   @@@@@  @@@@@  @@@@
//   @   @  @   @  @@   @@      @ @      @   @  @   @  @      @      @   @
//   @   @  @   @  @@   @@      @ @      @      @   @  @      @      @   @
//   @@@@   @@@@   @ @ @ @      @@        @@@   @@@@   @@@@@  @@@@@  @   @
//   @  @   @      @ @ @ @     @  @@         @  @      @      @      @   @
//   @  @   @      @  @  @     @  @      @   @  @      @      @      @   @
//   @   @  @      @  @  @      @@ @      @@@   @      @@@@@  @@@@@  @@@@
//
// Interrupt Service Routine called whenever the RPM input
// receives the rising edge of an incoming pulse.  Each ISR
// function must take no parameters and return nothing.
//
void ShaftPulseISR(void)
{
  m_ulMicros = micros();
  m_ulPulses++;
}
// This function should be called periodically to report the average
// shaft speed since the last call.  The suggested interval is about
// once or twice per second.  Testing on one Arduino UNO Rev3 board
// revealed about 0.2% error in RPM using the 60-Hz power line.
//
// returns: Scaled and converted Rvolutions per Minute
//
int ReadRPM()
{
  noInterrupts(); // disable changes by ISR during read
  unsigned long ulCurPulses = m_ulPulses;
  unsigned long ulCurMicros = m_ulMicros;
  interrupts();   // restore normal operation

  // RTC microseconds is always a multiple of 4 in ATmega328.
  // Scale elapsed time in uS by 1/4 without losing precision
  unsigned long ulPulses  =  ulCurPulses - m_ulPrevPulses;
  unsigned long ulMicros4 = (ulCurMicros - m_ulPrevMicros) >> 2;

  if ((m_Settings.nTachDiv == 0) || (ulPulses == 0) || (ulMicros4 == 0))
    return 0; // prevent division by 0

    // Shift current parameters into holding buffer for next time
  m_ulPrevPulses = ulCurPulses;
  m_ulPrevMicros = ulCurMicros;

  // Calculate shaft RPM from the pulse count and elapsed time.
  //
  //         # Pulses   10^6 uSec    1 Rev     60 sec
  //  RPM =  -------- * --------- * -------- * ------
  //          # uSec       Sec      N Pulses    min
  //
  // unsigned longs max out at 4,294,967,295 so we calculate
  // the equation above in a sequence to avoid overflow.
  // Add half the demoninator before final division to
  // round to the nearest whole number of RPM.
  // This easily supports up to 10K RPM at 4 pulses/rev.
  //
  // ---------- RPM CALIBRATION NOTE ----------
  // The frequency of the crystal driving the system clock on your Arduino
  // board will deviate slightly from the ideal value, affecting your
  // measured RPM.  We compensate for this here by adjusting the pulse
  // multiplier constant (MS_PER_MINUTE) by a delta (m_Settings.nTachCalib).
  //    RPM  = ulPulses * MS_PER_MINUTE / m_Settings.nTachDiv * 250;
  unsigned long ulMsPerMin = (unsigned long)((long)MS_PER_MINUTE + 
    (long)m_Settings.nTachCalib);

  // Calculate RPM using 250 denom since ulMicros4 is already scaled by 4
  unsigned long ulRPM  = ulPulses * ulMsPerMin / m_Settings.nTachDiv * 250;
  
  // ----------- DRIVE MECHANICAL RATIO NOTE -----------
  // If the tachometer sensor cannot be conveniently mounted to directly monitor
  // the speed of the shaft of interest, we can mount it in a more convenient place
  // as long as the mechanical drive ratio between the two is constant.  A common
  // situation is to monitor the motor shaft speed and compensate for the belt drive
  // ratio to determine the final quill speed.  We make this calculation ONLY when
  // the mechanical drive ratio is a value OTHER than 1-to-1.
  if (m_Settings.nTachScale != TACH_SCALE_DENOM) // not 1:1 ratio?
      ulMicros4  = ulMicros4 * TACH_SCALE_DENOM / m_Settings.nTachScale;
  
  ulRPM = (ulRPM + (ulMicros4 >> 1)) / ulMicros4; // round to nearest 0.5
  if (ulRPM > INT_MAX)           ulRPM = INT_MAX; // clamp at max int
  return ((int)ulRPM); 
}

//______________________________________________________________
//
// Calculate cutting speed in either SAE or SI units of measure
// from the specified shaft RPM and tool diameter.
//           SPEED = RPM * Diameter * pi
// We use the integer fraction 355/113 as a close approximation
// of pi.  Trivia: this fraction was used by Hewlett Packard around
// 1970 as an example in the docs for the famous HP-35 calculator.
//
// Returns: linear cutting speed in the current engineering units,
// as meters/minute (when m_Settings.bMetric=true), else feet/minute.
//
int CalcSpeed(int nRPM)
{
  // 32-bit unsigned long handles up to 40K RPM
  unsigned       long ulRPM = (unsigned long)nRPM;
  const unsigned long denom = (m_Settings.bMetric) ? 
  //
  // for SI units of measure, calc speed in meters/min:
  //
  //                 Rev   Diameter_mm     1 m      355
  // SPEED (m/min) = --- * ----------- * ------- *  ---
  //                 min                 1000 mm    113
  //
  (PI_DENOM * MILLIMETERS_PER_METER) :
  //
  // For SAE units of measure, calc speed in surface ft/min
  //
  //                Rev   Diameter_16ths in    1 ft    355
  // SPEED (SFPM) = --- * ----------------- * ----- *  ---
  //                min          16           12 in    113
  //
  (PI_DENOM * INCHES_PER_FOOT * SIXTEENTHS_PER_INCH);

  return (ulRPM * m_Settings.nDiameter * PI_NUMER / denom);
}

//_________________________________________________________________
//
// Format the tool/work diameter into a memory buffer for display.
// The buffer will be padded with spaces up to the specified length.
//
void FormatDiameter(
      char* pBuffer,   // destination buffer to receive string, returned to caller
      short nMinChars) // minimum (nominal) width of output in characters
{
  if (m_Settings.bMetric) // display dia in millimeters (very simple)
  {
    itoab(pBuffer, m_Settings.nDiameter, 3, RADIX_DEC); // show mm in 3-char field
    strcat(pBuffer, " mm");
  }
  else // display dia in fractional inches
  Sixteenths2Frac(pBuffer, m_Settings.nDiameter);

  strcat(pBuffer, pSpc(nMinChars - strlen(pBuffer))); // pad to specified length
}

//_________________________________________________________________
//
// Convert 16ths (of an inch) into an ASCII string for display
// as whole and fractional inches like "1-3/8"
//
char* Sixteenths2Frac(
      char* pBuffer, // destination buffer to receive string, returned to caller
      int   n16ths)  // numeric value to be converted
{
  int denom   = SIXTEENTHS_PER_INCH; // unreduced denominator
  int inches  = n16ths / denom;      // calc whole inches
  n16ths      = n16ths % denom;      // calc remaining 16ths
  *pBuffer    = '\0';                // make Buffer valid/empty string

  if (inches)                        // show whole inches (if any)
    itoab(pBuffer, inches, 0, RADIX_DEC);

  if (inches && n16ths)              // insert whole/frac separator?
    strcat(pBuffer, "-");

  if (n16ths)                        // show fractional inches (if any)
  {
    // If numerator is an even number, fraction can be reduced
    while(0 == (n16ths & B00000001)) // repeat until numerator is odd
    {
      n16ths = n16ths >> 1;
      denom  =  denom >> 1;
    }
    itoab(pEndOf(pBuffer), n16ths, 0, RADIX_DEC); // numerator
    strcat(pBuffer, "/");
    itoab(pEndOf(pBuffer),  denom, 0, RADIX_DEC); // denominator
  }
  strcat(pBuffer, "\"");                    // append inches symbol
  return pBuffer;
}

// ############################################################################
//
//   @   @ @@@@@ @  @     @ @@@@@ @  @@@@@   @@@
//   @   @   @   @  @     @   @   @  @      @   @
//   @   @   @   @  @     @   @   @  @      @
//   @   @   @   @  @     @   @   @  @@@@@   @@@
//   @   @   @   @  @     @   @   @  @          @
//   @   @   @   @  @     @   @   @  @      @   @
//    @@@    @   @  @@@@  @   @   @  @@@@@   @@@
//
//___________________________________________________________________
//
// This handler provides a simple way to report runtime errors
// via the LCD.  Calls to this routine will never return.
//
void FatalErr(char* pMsg, // text to be displayed before numeric argument
              int   arg)  // numeric argument displayed at end
{
  digitalWrite(PIN_PWM, LOW); // Motor OFF!
  LcdPrintLine(UPPER_LINE, "* Fatal Error * ");
  LcdPrintLine(LOWER_LINE, pMsg);
  m_lcd.print(arg);  
  m_lcd.print(pSpc(9));
  
  #ifdef DEBUG_DUMP //------------//
  Serial.print("FATAL ERROR: ");  //
  Serial.print(pMsg);             //
  Serial.println(arg);            //
  #endif // DEBUG_DUMP -----------//  
  
  while (true) { }; // loop forever here
}


// ############################################################################
//
//   @@@@        @@ @@                     @      @@@@        @              @
//   @   @       @  @                      @     @    @       @              @
//   @   @ @   @@@@@@@  @@@  @@@  @@@   @@@@     @    @ @   @@@@ @@@@  @   @@@@
//   @@@@  @   @ @  @  @   @ @   @   @ @   @     @    @ @   @ @  @   @ @   @ @
//   @   @ @   @ @  @  @@@@@ @   @@@@@ @   @     @    @ @   @ @  @   @ @   @ @
//   @   @ @   @ @  @  @     @   @     @   @     @    @ @   @ @  @   @ @   @ @
//   @@@@   @@@@ @  @   @@@@ @    @@@@  @@@@      @@@@   @@@@ @@ @@@@   @@@@ @@
//                                                               @
//                                                               @
//____________________________________________________________
//
// Return a pointer to an initialized string scratch buffer
// containing a specified number of spaces (0-16);
//
char* pGetStrBuffer(short nSpcs)
{
  char* pBuf = &m_szBuffer[0]; // Prepare empty buffer
  strcpy(pBuf, pSpc(nSpcs)); // Init with specified indent
  return (pBuf);
}
//____________________________________________________________
//
// return a pointer to the END of the string at pBuffer.
// This is helpful in the concatenation of strings.
//
char* pEndOf(char* pBuffer)
{
  return (pBuffer + strlen(pBuffer));
}
//____________________________________________________________
//
// Return a string containing the specified number of spaces
//
char* pSpc(int nSpcs) // number of spaces
{
  const int nMax = sizeof(SPACES16)-1; // exclude NUL terminator
  nSpcs = min(nSpcs, nMax);
  return (char*)(&SPACES16[nMax-nSpcs]);  
}
//___________________________________________________________
//
// Append spaces onto the string in the specified buffer
// until it will fill an entire line of the LCD, thus
// erasing any prior displayed characters at these locations.
// This buffer is then displayed on the specified LCD line.
//
void LcdPrintBuffer(int line, char* pBuffer)
{
  int nLen = strlen(pBuffer);
  if (nLen < LCDCOLS)
    strcpy(pBuffer+nLen, pSpc(LCDCOLS-nLen));
  m_lcd.setCursor(0, line);
  m_lcd.print(pBuffer);
}

//____________________________________________________________
//
// Clear the line if pMsg is NULL
//
void LcdPrintLine(int line, const char* pMsg)
{
  m_lcd.setCursor(0, line);

  if (pMsg)  m_lcd.print((char*)pMsg);
  else       m_lcd.print((char*)&SPACES16);
}

/*____________________________________________________________
 // This routine is commented out, but left as example code
 //
 // Display an unsigned fixed-point value in the caller's buffer.
 // The value to be displayed is an unsigned long integer scaled by
 // the base-10 antilog of the value nDigits.  For example, the 
 // nValue=12345 and nDigits=3 would produce 12.345.
 //
 char*  fixtoa(  // Convert a fixed point numeric value to an ASCII string
 char*         pBuffer, // buffer to be overwritten, returned to caller
 unsigned long nValue,  // numeric value to be converted, * 10^nDigits
 short         nWidth,  // minimum width to output before decimal, leading 0's if negative
 short         nDigits) // # digits past decimal
 {
 unsigned long nDiv = 1;
 for (short nDig=0; nDig<nDigits; nDig++)
 nDiv *= 10;
 
 int nBefore = (int)(nValue / nDiv);
 int nAfter  = (int)(nValue % nDiv);
 
 itoab(pBuffer, nBefore, nWidth, RADIX_DEC);
 strcat(pBuffer, ".");
 itoab(pEndOf(pBuffer), nAfter, nDigits, RADIX_DEC);
 return pBuffer;  
 }
 */

//____________________________________________________________
//
// Display the specified ADC value (raw Counts) in Volts and
// append onto the buffer as an 8-char string like " 4.567 V",
// then display the results on the lower line of the LCD.
// The caller should prepare the buffer with a suitable prefix.
//
void DisplayAdcVolts(char* pBuf, unsigned int nCounts)
{
  // Convert A/D raw counts into DC millivolts using integer math.
  // We use 32-bit varbs to prevent overflow (max value 5000*1024)
  // and round to the nearest whole millivolt.
  int nValue  = (int)(((unsigned long)nCounts*ADC_FS_MV + ADC_FULL_SCALE/2) / ADC_FULL_SCALE);

  // Format nValue as a float, divided by 1000 (Show 1234 as " 1.234 V")
  // We could format the value into the string buffer and then cram
  // in a decimal point, but this approach uses existing routines.
  itoab(pEndOf(pBuf), nValue/1000,  2, RADIX_DEC); // digits before decimal
  strcat(pBuf, ".");
  itoab(pEndOf(pBuf), nValue%1000, -3, RADIX_DEC); // 3 digits past decimal
  strcat(pBuf, " V");

  LcdPrintBuffer(LOWER_LINE, pBuf); // xfer buffer to LCD lower line
}

//____________________________________________________________
//
// This function is a variant of the itoa (integer-to-ascii-string)
// function in the C "stdlib" library.  This function supports any
// radix, although base 10 and 16 are the only ones we will use.
// This function differs from the itoa() baseline implementation
// mainly in its ability to maintain a minimum field width by padding
// the output with leading spaces, or zeros if nWidth is negative.
//
// Examples:
// char   buff[8]; // buffer needs room for output incl NUL terminator
// itoab(&buff[0], 123,  0, 10); // prints "123" like most print() funcs
// itoab(&buff[0], 123,  4, 10); // prints " 123" (4 char width)
// itoab(&buff[0], 123, -4, 16); // prints hex w/ leading 0 as "007B"
//
char* itoab(          // Convert a numeric value to an ASCII string
      char*  pBuffer, // buffer to be overwritten, returned to caller
      int    nValue,  // numeric value to be converted
      short  nWidth,  // minimum width to output, leading 0's if negative
      short  nRadix)  // number base, 2-16
{
  const char digits[] = "0123456789ABCDEF"; // letter "F" handles <= base 16
  boolean    bSign    = (nValue < 0);       // value needs "-" sign prefix?
  boolean    bZero    = (nWidth < 0);       // leading zeros requested?
  char       cPad     =  bZero ? '0' : ' ';
  int        j, i=0;

  if (bSign)  nValue = -nValue; // use absolute values
  if (bZero)  nWidth = -nWidth;

  do  // most of the conversion happens here
  {   // output digits in reverse sequence (LSD first)
    *(pBuffer + i++) = digits[nValue % nRadix];
  }   
  while (nValue /= nRadix);

  if (bSign) // insert a '-' sign prefix for negative numbers
  {
    j = i++;
    *(pBuffer + j) = '-'; // insert sign
  }
  else
    j = i - 1;

  while (i < nWidth)          // caller specified a min width?
    *(pBuffer + i++) = cPad;  // i points to next available

  if ((--i != j) && bSign && bZero)  // format like "-000123"?
  {
    *(pBuffer + i) = *(pBuffer + j); // get the sign
    *(pBuffer + j) = cPad;           // fill in over it
  }

  *(pBuffer + ++i) = '\0';       // strings must be NUL terminated

  for (i--, j=0; j<i; j++, i--)  // repair (reverse) char seq
  {
    char cTmp      = *(pBuffer + j);
    *(pBuffer + j) = *(pBuffer + i);
    *(pBuffer + i) =   cTmp;
  }
  return pBuffer;
}
//    *****************************************
//      End of Sketch, about 2800 lines total
//    *****************************************













